#! python
"""
script: aprcompiler
Author: Ernesto Posse
Created on: July 4, 2006
Last update: July 31, 2006

Description:
    This script compiles a .cfg file into a python representation.
    The standard command-line usage is:
    
        cfgcompiler.py [options] <file>
    
    where <file> must have a .cfg extension, and [options] are the following:
      -a         [Default] Use ASCII format for the generated pkl file. Larger file, slower to load, but portable.
      -b         Use binary format for the generated pkl file. Smaller file, faster to load, but may not be portable.
      -d <dir>   Save the generated package in the given target directory <dir>.
      -m         [Default] Minimal save: saves only the basic python representation file. It is enough when using build_parser. 
      -f         Full save: saves the generated parser and actions. Necessary when using load_parser instead of build_parser.
      -v         Verbose; shows various internal messages.
    
    If the file name is of the form <some_name>.apr then the following will be 
    generated in the same directory where the file is located:
    
    - <some_name>.py:   A file containing a python representation of the grammar,
    - <some_name>_cfg:  A directory containing a python package with that name 
                        and which contains the following files:
        - __init__.py: .................... A file to identify the <some_name>_cfg 
                                            directory as a python package.
        - <some_name>_cfg_ll1_actions.py: . A python module containing the source 
                                            code for the actions associated with 
                                            grammar rules.
        - <some_name>_cfg_ll1_wan.py: ..... A python module containing the normalized 
                                            form of the python representation of the 
                                            grammar.
        - <some_name>.pkl: ................ A python pickled file containing the parser 
                                            object itself (including the lexer and the 
                                            CFG object.)
    
Typical usage:
    1) Create a source apr file, for example: mylang.apr
    
        separators
            colon ":"
        identifiers
            id
        numbers 
            age
        strings
            name
        start 
            RECORD
        rules
        RECORD -> name colon age : "('by name', $1, $3)"
                | id colon age : "('by id', $1, $3)"
    
    2) On the command line type:
        
        apr mylang.apr
    
    3) To use the generated parser in a python module, do:
    
        from aperiot.parsergen import load_parser
        parser = load_parser('mylang')
        text = '"Ernesto Posse":1329'
        outcome = parser.parse(text)
        name, age = outcome

    Note: the generated files, as well as the aperiot package, must be in the python path (i.e. must appear in sys.path.)
"""

import sys
sys.path += ['.','..', '../src']

import getopt
import os
import os.path
import posse_tools.path_setup
posse_tools.path_setup.setup_paths(required_modules=['llparser'], strict=False, verbosity=0)

import aperiot
pkg_path = os.path.abspath(aperiot.__path__[0])
sys.path += [pkg_path]

from aperiot.parsergen import *
from aperiot.llparser import ParsingException
from posse_tools.utils import *
from aperiot.metacfg_wat_cfg import *

#~ meta_cfg_parser = 'metacfg_wat'
meta_cfg_parser = 'metacfg_bootstrap'
version = aperiot.version

def compile_cfg(parser, dir, source, pkl_format, full_save, verbose):
    text = file(source, 'r')
    if verbose:
        print_time("* Parsing source cfg file '"+source+"'")
    grammar = parser.parse(text)
    if verbose:
        print_time("* Source cfg file parsed")
    text.close()
    generate_pyrepr(dir, source, grammar, verbose)
    if full_save:
        imports = grammar[0]
        cfg_description_module, package_name, dir = get_module_pkginfo(source)
        if verbose:
            print_time("* Building parser")
        new_parser = build_parser(cfg_description_module, verbose=verbose)
        if verbose:
            print_time("* Parser built")
            print_time("* Saving generated parser")
        save_parser(new_parser, 
                    dir,
                    cfg_description_module, 
                    pickle_format=pkl_format,
                    verbose=verbose)
        if verbose:
            print_time("* Generated parser saved")

def get_module_pkginfo(full_source_name):
    dir, source = os.path.split(full_source_name)
    dir = os.path.abspath(dir)
    cfg_description_module, ext = os.path.splitext(source)
    #~ base_dir, package_name = os.path.split(dir)
    package_name = cfg_description_module + '_cfg'
    if package_name == '':
        package_name = '__main__'
    return cfg_description_module, package_name, dir

def generate_pyrepr(dir, source, grammar, verbose=False):
    new_file_name = make_new_file_name(source)
    if verbose:
        print_time("* Generating python representation")
    generated_text = generate_text(grammar)
    if verbose:
        print_time("* Python representation generated")
    cfg_description_module, package_name, dir = get_module_pkginfo(source)
    package_path = CFG.get_package_path(package_name)
    if package_path is None:
        if verbose:
            print_time("* Creating new package: '" + package_name + "' in '" + str(os.getcwd()) + "'") 
        package_path = CFG.create_package(package_name, dir)
    elif verbose:
        print_time("* Package '" + package_path + "' already exists. Saving new modules there")
    new_file_name = os.path.join(package_path, new_file_name)
    new_file = file(new_file_name, 'w')
    new_file.write(generated_text)
    new_file.close()
    if verbose:
        print_time("* Python representation saved in file '"+new_file_name+"'")

def extract_terminals(terminals):
    try:
        keywords    = terminals['keywords']
    except KeyError:
        keywords    = []
    try:
        operators   = terminals['operators']
    except KeyError:
        operators   = []
    try:
        separators  = terminals['separators']
    except KeyError:
        separators  = []
    try:
        brackets    = terminals['brackets']
    except KeyError:
        brackets    = []
    try:
        identifiers = terminals['identifiers']
    except KeyError:
        identifiers = []
    try:
        numbers     = terminals['numbers']
    except KeyError:
        numbers     = []
    try:
        strings     = terminals['strings']
    except KeyError:
        strings     = []
    return (keywords, operators, separators, brackets, identifiers, numbers, strings)
    
def generate_text(grammar):
    imports, useindents, usenewlines, terminals, start, rules = grammar
    keywords, operators, separators, brackets, identifiers, numbers, strings = extract_terminals(terminals)
    rules_text = generate_rules_text(rules)
    generated_text = "\n"
    generated_text += "from aperiot.lexer import *\n\n"
    generated_text += "imports = [" + ', '.join(["'" + item + "'" for item in imports]) + "]\n"
    generated_text += "\n"
    generated_text += "useindents = "+str(useindents)+"\n"
    generated_text += "usenewlines = "+str(usenewlines)+"\n"
    generated_text += "\n"
    generated_text += "# Terminal symbols\n\n"
    generated_text += generate_terminal_defs(useindents, usenewlines, keywords, 
                                                operators, separators, brackets, 
                                                identifiers, numbers, strings) 
    generated_text += "# Non-terminal symbols\n\n"
    generated_text += generate_nonterminal_defs(rules)
    generated_text += "# Start symbol\n\n"
    generated_text += "start = '" + start + "'\n\n"
    generated_text += "# Rules\n\n"
    generated_text += "rules = \\\n" + indent(rules_text, 4) + "\n"
    generated_text += "\n"
    return generated_text

def generate_terminal_defs(useindents, usenewlines, keywords, 
                            operators, separators, brackets, 
                            identifiers, numbers, strings):
    terminal_defs = ""
    if useindents:
        terminal_defs += "indent  = Indent\n"
        terminal_defs += "dedent  = Dedent\n"
    if usenewlines:
        terminal_defs += "newline = Newline\n"
    if len(keywords) > 0:
        keyword_strings = [k.string for k in keywords]
        size1 = max(len(k) for k in keyword_strings)
        keyword_symbols = [k.symbolic_name for k in keywords]
        size2 = max(len(k) for k in keyword_symbols)
        size = max(size1, size2)
        for keyword in keywords:
            terminal_defs += fit(keyword.symbolic_name, size) + " = " + str(keyword) + "\n"
    if len(operators) > 0:
        operator_strings = [o.symbolic_name for o in operators]
        size = max(len(o) for o in operator_strings)
        for operator in operators:
            terminal_defs += fit(operator.symbolic_name, size) + " = " + str(operator) + "\n"
    if len(separators) > 0:
        separator_strings = [o.symbolic_name for o in separators]
        size = max(len(o) for o in separator_strings)
        for separator in separators:
            terminal_defs += fit(separator.symbolic_name, size) + " = " + str(separator) + "\n"
    if len(brackets) > 0:
        bracket_strings = [o.symbolic_name for o in brackets]
        size = max(len(o) for o in bracket_strings)
        for bracket in brackets:
            terminal_defs += fit(bracket.symbolic_name, size) + " = " + str(bracket) + "\n"
    if len(identifiers) > 0:
        identifiers_strings = [k.string for k in identifiers]
        size = max(len(k) for k in identifiers_strings)
        for identifier in identifiers:
            terminal_defs += fit(identifier.string, size) + " = Identifier\n"
    if len(numbers) > 0:
        number_strings = [o.string for o in numbers]
        size = max(len(o) for o in number_strings)
        for number in numbers:
            terminal_defs += fit(number.string, size) + " = NumberLiteral\n"
    if len(strings) > 0:
        string_strings = [o.string for o in strings]
        size = max(len(o) for o in string_strings)
        for string in strings:
            terminal_defs += fit(string.string, size) + " = StringLiteral\n"
    terminal_defs += "\n\n"
    return terminal_defs 

def generate_nonterminal_defs(rules):
    nonterminal_defs = ""
    size = max(len(key) for key in rules) 
    for key in rules:
        nonterminal_defs += fit(key, size) + " = '" + key + "'\n"
    nonterminal_defs += "\n\n"
    return nonterminal_defs 
    
def generate_rules_text(rules):
    rules_text = "{\n"
    for key in rules:
        rules_text += " " * 4 + key + ": \n"
        rules_text += " " * 8 + "[\n"
        for word, action in rules[key]:
            symbols = ', '.join([str(s) for s in word])
            if action == '':
                action = 'None'
            rules_text += " " * 12 + "([" + symbols + "], '" + str(action) + "'), \n"
        rules_text += " " * 8 + "],\n"
    rules_text += "}\n"
    return rules_text

def make_new_file_name(full_source_name, overwrite=True):
    base_path, source = os.path.split(full_source_name)
    root, ext = os.path.splitext(source)
    new_name = root + '.py'
    if not overwrite and os.path.exists(new_name):
        raise CFGCompilerException("The file '" + new_name + "' already exists. Please use a different name for your cfg file, or remove the .py file")
    return new_name

def usage():
    print "usage: cfgcompiler.py [options] <file>"
    print "where <file> is a .apr file,"
    print "and [options] is a list of one of the following options:"
    print "  -a         [Default] Use ASCII format for the generated pkl file. \n" \
        +  "                Larger file, slower to load, but portable."
    print "  -b         Use binary format for the generated pkl file. \n" \
        +  "                Smaller file, faster to load, but may not be portable."
    print "  -m         [Default] Minimal save: saves only the basic python representation \n" \
        +  "                file. It is enough when using build_parser. "
    print "  -f         Full save: saves the generated parser and actions. Necessary when \n" \
        +  "                using load_parser instead of build_parser."
    print "  -d <dir>   Save the generated package in the given target directory <dir>."
    print "  -t         Use the hard-coded meta-grammar instead of the bootstrapped \n" \
        +  "                meta-grammar."
    print "  -v         Verbose; shows various internal messages."
    print "  -V         Print version number."
    print "  -h, --help Print this message."

def print_version():
    print version

def main():
    try:
        options, args = getopt.getopt(sys.argv[1:], 'abd:mfvhtV',['help'])
        verbose = False
        full_save = False
        ASCII      = 0
        BINARY     = 2
        OLD_BINARY = 1
        pickle_format = ASCII
        target_dir = os.getcwd()
        global meta_cfg_parser
        for option, arg in options:
            if option == '-v':
                verbose = True
            if option == '-a':
                pickle_format = ASCII
            if option == '-b':
                pickle_format = BINARY
            if option == '-d':
                target_dir = os.path.abspath(arg)
            if option == '-m':
                full_save = False
            if option == '-f':
                full_save = True
            if option == '-t':
                meta_cfg_parser = 'metacfg_wat'
            if option == '-V':
                print_version()
                sys.exit(2)
            if option in ('-h', '--help'):
                usage()
                sys.exit(2)
        if len(args) != 1:
            raise getopt.GetoptError("No file given")
        full_source_name = args[0]
        #~ base_dir, source = os.path.split(full_source_name)
        source = os.path.abspath(full_source_name)
        if verbose:
            print_time("* Loading CFG parser")
        parser = load_parser(meta_cfg_parser, verbose)
        if verbose:
            print_time("* CFG parser loaded")
        current_dir = os.getcwd()
        os.chdir(target_dir)
        compile_cfg(parser, target_dir, source, pickle_format, full_save, verbose)
        os.chdir(current_dir)
        if verbose:
            print_time("* CFG compilation completed")
    except getopt.GetoptError, e:
        print str(e)
        usage()
        sys.exit(2)
    except IOError, e:
        print_time('* (IOError) '+str(e))
    except ParsingException, e:
        print_time('* (ParsingException) '+str(e))
    except CFGCompilerException, e:
        print_time('* (CompilerException) '+str(e))

class CFGCompilerException(Exception):
    pass

if __name__ == "__main__":
    main()
