"""
Module: cfgs
Author: Ernesto Posse
Created on: April 13, 2006
Last update: Jan 22, 2007

Description:
  Context Free Grammars
"""


from lexer import *
from types import *
from plambda.plambda_ast import *
from posse_tools.utils import *

import copy 
import os
import os.path
import inspect
import string

import pprint
pp = pprint.PrettyPrinter(indent=2)

class CFG:
    
    def __init__(self, start, composite_rule_table, modules_to_import=[]):
        self.start = NonTerminal(start)
        self.rules = []
        self.actions_table = {}
        self.original_actions = {}
        self.all_names = []
        self.normalized = False
        self.modules_to_import = modules_to_import
        if CFG.is_valid_dict_repr(composite_rule_table):
            self.composite_rule_table_with_templates = composite_rule_table
            self.composite_rule_table_with_actions = \
                self.__generate_plamterms_from_action_templates(composite_rule_table)
            self.__generate_flat_rule_table(self.composite_rule_table_with_actions)
            self.__populate_actions_table()
            self.generate_plam_execution_environment(self.modules_to_import)
        else:
            raise CFGException('The composite rule table is invalid.')
        self.name_gen = NameGenerator(base_name='S', names_to_avoid=self.all_names, upper=True)
        self.ind_gen = IndexedNameGenerator(base_name='S')
    
    """
    Public instance methods
    """
    
    # Public instance accessors, selectors and extractors
    
    def __getitem__(self, rule_number):
        return self.rules[rule_number]
    
    def __getstate__(self):
        obj_dict = copy.copy(self.__dict__)
        del obj_dict['environment']
        return obj_dict
    
    def __setstate__(self, obj_dict):
        self.__dict__.update(obj_dict)
        package_name = obj_dict['package_name']
        normalized   = obj_dict['normalized']
        if not normalized:
            module = 'original'
        else:
            module = 'll1'
        module_name = package_name + '_' + module + '_actions'
        #~ exec "from " + module_name + " import *"
        #~ print "importing "
        #~ print self.modules_to_import
        self.generate_plam_execution_environment(self.modules_to_import)
    
    def __str__(self):
        s = ""
        for r in ((str(i) + ": " + str(self.rules[i])) for i in range(len(self.rules))):
            s += r + '\n'
        return s
    
    def dict_repr_str(self, max_sym_size=20):
        outcome = "\n"
        outcome += "# Non-terminal symbol names\n\n"
        for key in self.composite_rule_table_with_actions:
            outcome += str(key) + " = '" + str(key) + "'\n"
        outcome += "\n"
        outcome += 'rules = \\\n'
        base = 2
        outcome += indent("{", base)
        outcome += "\n"
        for key in self.composite_rule_table_with_actions:
            outcome += indent(str(key) + ": ", base * 2)
            outcome += "\n"
            outcome += indent("[", base * 3)
            outcome += "\n"
            for word, act in self.composite_rule_table_with_actions[key]:
                sym_list = ', '.join([str(sym) for sym in word])
                if type(act) is str:
                    act_str = act
                elif callable(act):
                    act_str = act.__name__
                outcome += indent("([" + sym_list + "], '" + act_str + "'),", base * 4)
                outcome += "\n"
            outcome += indent("], ", base * 3)
            outcome += "\n"
        outcome += indent("}",base)
        outcome += "\n\n"
        outcome += "start = " + str(self.start.sym) + "\n"
        outcome += "\n\n"
        return outcome
    
    #~ def generate_actions_file(self, module_name=None, additional_actions=None, modules_to_import=None):
        #~ self.__save_actions_file(module_name, additional_actions, modules_to_import)
    
    #~ def generate_python_repr(self, module_name=None):
        #~ self.__save_python_repr(module_name)
    
    def nice_str(self, max_sym_size=20):
        outcome = ''
        for key in self.composite_rule_table_with_actions:
            outcome += str(key) + ' -> '
            first = True
            level = 0
            bar = ''
            for word, act in self.composite_rule_table_with_actions[key]:
                if first:
                    first = False
                else:
                    level = 4
                    bar = '| '
                sym_list = ''
                if len(word) == 0:
                    sym_list = 'empty'
                else:
                    sym_list = ' '.join([short_str(sym, max_sym_size) for sym in word])
                sym_list = bar + sym_list
                if callable(act):
                    act_str = act.__name__
                else:
                    act_str = str(act)
                outcome += indent(sym_list, level) + ' : ' + act_str + '\n'
            outcome += '\n' 
        outcome += 'start\n' + str(self.start.sym) 
        outcome += '\n' 
        return outcome
    
    def save_package(self, 
                        package_name, 
                        module='original', 
                        modules_to_import=[], 
                        additional_actions=[],
                        verbose=False):
        """
        The modules parameter should be 'original' or 'll1'
        """
        self.package_name = package_name
        package_path = self.get_package_path(package_name)
        if package_path is None:
            package_path = self.create_package(package_name)
            if verbose:
                print_time("* Creating new package: '" + package_path + "'") 
        cfg_module_name     = package_name + '_' + module + '_cfg_wan'
        actions_module_name = package_name + '_' + module + '_actions'
        self.__save_dict_repr_module(package_path, cfg_module_name, verbose)
        self.__save_actions_module(package_path, actions_module_name, modules_to_import, additional_actions, verbose)
    
    # Public instance mutators
    
    def generate_execution_environment(self, modules_to_import=None):
        if modules_to_import is None:
            modules_to_import = self.modules_to_import
        env = {}
        for module in modules_to_import:
            exec "from " + module + " import *" in env
        #~ print self.actions_table.keys()
        for action_name in self.actions_table:
            action = self.actions_table[action_name]
            if isinstance(action, RuleAction):
                print "rule action "+str(action.__class__)
                print str(action)
                exec str(action) in env
            elif type(action) is FunctionType:
                #~ print "action is a function"
                env[action_name] = action
        #~ print_dict(env,a=20)
        for action_name in self.actions_table:
            action = self.actions_table[action_name]
            action.environment = env
            action.function = env[action_name]
        #~ print 'adding primitives'
        #~ print __builtins__.__dict__
        #~ print type(__builtins__.__dict__)
        #~ print __builtins__.__dict__.keys()
        for primitive in __builtins__.__dict__:
            #~ print 'adding primitive:',primitive
            env[primitive] = __builtins__.__dict__[primitive]
        self.environment = env
        
    def generate_plam_execution_environment(self, modules_to_import=None):
        if modules_to_import is None:
            modules_to_import = self.modules_to_import
        env = {}
        for module in modules_to_import:
            exec "from " + module + " import *" in env
        for action_name in self.actions_table:
            action = self.actions_table[action_name]
            env[action_name] = action
            action.environment = env
        #~ print 'adding primitives'
        #~ print __builtins__
        #~ print type(__builtins__)
        #~ print __builtins__.keys()
        for primitive in __builtins__:
            #~ print 'adding primitive:',primitive
            env[primitive] = __builtins__[primitive]
        self.environment = env
        #~ print '\nself.environment:\n'
        #~ print_dict(self.environment)
        #~ print "\nconstants\n"
        #~ for c in constants:
            #~ print str(c)
        #~ print
        
    def inline_rules(self, update=True):
        new_table, new_actions = self.__compute_inline_expansion()
        #~ print "\nafter inlining actions table\n"
        #~ print_dict(self.actions_table, items='kt')
        #~ print "\nnew actions:\n"
        #~ for a in new_actions:
            #~ print a
        self.composite_rule_table_with_actions = new_table
        self.__generate_flat_rule_table(self.composite_rule_table_with_actions)
        for action in new_actions:
            self.actions_table[action.name] = action
        #~ self.__collect_actions_garbage()
        #~ print_dict(self.actions_table, a=25)
        
    def normalize(self):
        #~ self.print_tables('START------------------------------------------------------------------------------')
        self.inline_rules()
        #~ print "-- finished firts inlining"
        self.__simplify_actions()
        #~ self.print_tables('PHASE 1: after first inlining --------------------------------------------------------')
        self.__collect_actions_garbage()
        #~ self.print_tables('PHASE 1\': after first gc -------------------------------------------------------------')
        self.transform_to_ll1()
        #~ print "-- finished transform to ll1"
        self.__simplify_actions()
        #~ self.print_tables('PHASE 2: after transform to ll1 -----------------------------------------------------')
        self.__collect_actions_garbage()
        #~ self.print_tables('PHASE 2\': after second gc-------------------------------------------------------------')
        self.inline_rules()
        #~ print "-- finished second inlining"
        self.__simplify_actions()
        #~ self.print_tables('PHASE 3-: after second inlining -----------------------------------------------------')
        self.__collect_actions_garbage()
        #~ self.print_tables('PHASE 3\': after third gc -------------------------------------------------------------')
        self.generate_plam_execution_environment()
        #~ self.print_tables('FINAL')
        self.normalized = True

    def print_tables(self, msg=''):
        self.__count_action_references()
        print "\n*** Tables **************************************************************************************\n"
        print msg
        #~ print "\n-- rule table with actions\n"
        #~ print_dict(self.composite_rule_table_with_actions, items='kt')
        print "\n-- rules -- \n"
        print self.nice_str()
        print "\n-- actions table\n"
        print_dict(self.actions_table, value_width=100, value_delim='\n', items='kt')
        print "\n-- free vars\n"
        for a in self.actions_table:
            act = self.actions_table[a]
            print 'action:',act.name
            print '  occurrences:',act.ocurrences
            print '  crosslinks:',act.crosslinks
            print '  free vars:',act.term.freevars()
            print '  term:'
            print indent(str(act),4)
        print "\n-- env size\n"
        print len(self.environment)
        print "\n-- env\n"
        print_dict(self.environment, value_width=10,items='kt')
        print "\n-- key diff\n"
        print str(set(self.environment) - set(self.actions_table))
        print "\n*** End of tables **************************************************************************************\n"

    def transform_to_ll1(self):
        new_table, new_actions = self.__compute_ll1()
        self.composite_rule_table_with_actions = new_table
        self.__generate_flat_rule_table(self.composite_rule_table_with_actions)
        for action in new_actions:
            self.actions_table[action.name] = action
        #~ self.__collect_actions_garbage()

    """
    Public static methods
    """

    @staticmethod
    def create_package(package_name, dir=None):
        if dir is None:
            path = os.getcwd()
        else:
            path = os.path.abspath(dir)
        package_path = os.path.join(path, package_name)
        os.mkdir(package_path)
        package_init_file = os.path.join(package_path, '__init__.py')
        new_init_file = file(package_init_file, 'w')
        new_init_file.writelines(['# CFG package: ' + package_name])
        new_init_file.close()
        return package_path

    @staticmethod
    def get_package_path(package_name):
        temp_table = {}
        try:
            #~ import os
            #~ print 'epp',os.getcwd()
            #~ print 'epp:.',os.listdir('.')
            #~ print 'epp:p',os.listdir(package_name)
            #~ print 'epp',package_name
            exec "import " + package_name in temp_table
            #~ print 'epp','import succeeded'
            package_path = eval(package_name, temp_table).__path__[0]
            #~ print 'epp',package_path 
        except ImportError:
            #~ print 'epp','import failed'
            package_path = None
        return package_path

    @staticmethod
    def is_valid_dict_repr(object):
        #~ if type(object) is dict and all(type(key) is str for key in object.keys()) \
            #~ and all(type(object[key]) is list for key in object.keys()):
        if type(object) is not dict:
            return False
        else:
            for key in object:
                if type(key) is not str or type(object[key]) is not list:
                    return False
                for item in object[key]:
                    if type(item) is tuple and \
                    ( \
                        len(item) != 2 \
                        or not CFG.is_valid_word(item[0], object)\
                        or (type(item[1]) is not str and not callable(item[1])) \
                        or (callable(item[1]) and sum(len(i or []) for i in getargspec(item[1])) != len(item[0])) \
                    ):
                        return False
                    elif type(item) is list and not CFG.is_valid_word(item, object):
                        return False
        return True

    @staticmethod
    def is_valid_word(word, dict_repr):
        if type(word) is not list:
            return False
        return all(type(item) is str and item in dict_repr.keys() \
                    or type(item) is not str and \
                    (isinstance(item, Token) or issubclass(item, Token)) \
                    for item in word)
    
    """
    Private instance methods
    """
    
    # Private accessors, selectors and extractors
        
    def __compute_inline_expansion(self):
        #~ print "\n--** __compute_inline_expansion \n"
        original = self.composite_rule_table_with_actions
        original_actions = self.actions_table.keys()
        new_dict = original
        previous = {}
        new_actions_list = []
        while new_dict != previous:
            #~ print "NEW PAS **************************************************************"
            #~ print
            #~ print 'new_dict\n'
            #~ pp.pprint(new_dict)
            #~ print
            #~ print 'previous\n'
            #~ pp.pprint(previous)
            previous = self.__copy_composite_rule_table_with_actions(new_dict)
            #~ previous = copy.deepcopy(new_dict)
            rules_marked_for_elimination = []
            actions_marked_for_elimination = []
            for non_terminal in new_dict:
                #~ print 'Trying non-terminal:',non_terminal
                if non_terminal != self.start.sym and len(new_dict[non_terminal]) == 1:
                    #~ print '  exactly 1 alt'
                    rules_marked_for_elimination += [non_terminal]
                    expansion_word = new_dict[non_terminal][0][0]
                    expansion_action = new_dict[non_terminal][0][1]
                    #~ actions_marked_for_elimination += [expansion_action]
                    for other in new_dict:
                        if other != non_terminal:
                            #~ print '    replacing '+non_terminal+' in rules for '+other
                            #~ print '      before replacing:',str([str(a) for a in new_dict[other]])
                            new_alt_list = []
                            for w, a in new_dict[other]:
                                new_alt = []
                                matches = []
                                pos = 0
                                for sym in w:
                                    if sym == non_terminal:
                                        new_alt += expansion_word
                                        matches += [pos]
                                    else:
                                        new_alt += [sym]
                                    pos += 1
                                #~ print '        to expand:'
                                #~ print '          '+str(other) + '->' + str([str(s) for s in w])
                                #~ print '        with:'
                                #~ print '          '+str(non_terminal) + '->' + str([str(s) for s in expansion_word])
                                actions_marked_for_elimination += [a]
                                if len(matches) > 0:
                                    new_action = self.__make_new_expansion_plam_action(a,expansion_action,matches)
                                    #~ print '        new action:'
                                    #~ print indent(str(new_action),10)
                                else:
                                    #~ print '        no matches'
                                    new_action = a
                                new_actions_list += [new_action]
                                new_alt_list += [(new_alt, new_action)]
                            new_dict[other] = new_alt_list
                            #~ print '      after replacing:',str([str(a) for a in new_dict[other]])
            for non_terminal in rules_marked_for_elimination:
                del new_dict[non_terminal]
            #~ for action in actions_marked_for_elimination:
                #~ if action.name not in original_actions:
                    #~ print 'action to delete:',action.name
                    #~ del self.actions_table[action.name]
                
            #~ print "END PAS **************************************************************"
        #~ print 'Result\n'
        #~ pp.pprint(new_dict)
        return new_dict, new_actions_list
        
    def __compute_ll1(self):
        original = self.composite_rule_table_with_actions
        #~ print "\n--** __compute_ll1\n"
        #~ print_dict(original)
        #~ print "\n-- actions table\n"
        #~ print_dict(self.actions_table, value_width=10)
        #~ print "\n-- env\n"
        #~ print_dict(self.environment, value_width=10)
        #~ print "\n-- key diff\n"
        #~ print str(set(self.environment) - set(self.actions_table))
        name_gen = NameGenerator(base_name='S', names_to_avoid=original, upper=True)
        name_gen = self.name_gen
        pair_action = CFG.__make_pair_plam_action()
        new_actions_list = [pair_action]
        new_dict = {}
        remaining = original
        while remaining != {}:
            #~ print "NEW PASS ************************************************************************"
            new_remaining = {}
            for non_terminal in remaining:
                #~ print "  Trying non-terminal:",str(non_terminal)
                if non_terminal not in new_dict:
                    #~ print "    It wasn't in new_dict"
                    new_dict[non_terminal] = []
                i = 0
                for word, action in remaining[non_terminal]:
                    i += 1
                    #~ print "    Trying alternative word:",str([str(s) for s in word])
                    if word == []:
                        #~ print "      Empty word"
                        #~ new_action = self.__make_new_empty_word_action(non_terminal.lower() + '_action_' + str(i))
                        new_dict[non_terminal] += [([], action)]
                    else:
                        #~ print "      Non-empty-word"
                        first = word[0]
                        #~ new_firsts = [w[0] for w, a in new_dict[non_terminal]]
                        first_is_already_in_new_firsts, continuation = False, None
                        for w, a in new_dict[non_terminal]:
                            #~ print "        checking if it is in the new_firsts:",str(w)
                            if len(w) > 0 and first == w[0]:
                                first_is_already_in_new_firsts, continuation = True, w[1]
                                break
                        if first_is_already_in_new_firsts:
                            #~ print "        first is already in new_dict["+str(non_terminal)+"]"
                            #~ new_action_name = continuation + '_action_' + str(len(new_remaining[continuation]))
                            #~ new_action_name = name_gen.get_new_name(names_to_avoid = remaining.keys() + new_remaining.keys(), upper = True)
                            new_action_name = continuation + '_' + str(len(new_remaining[continuation]))
                            new_action = self.__make_new_symbol_plam_action(new_action_name, action.name, len(word) - 1)
                            new_remaining[continuation] += [(word[1:], new_action)]
                            new_actions_list += [new_action]
                        else:
                            #~ print "        first is not in new_dict["+str(non_terminal)+"]"
                            new_non_terminal = name_gen.get_new_name(non_terminal, remaining.keys() + new_remaining.keys(), True)
                            #~ new_non_terminal = non_terminal + '_0'
                            #~ new_non_terminal = name_gen.get_new_name(names_to_avoid = remaining.keys() + new_remaining.keys(), upper = True)
                            new_sym_action = self.__make_new_symbol_plam_action(new_non_terminal, action.name, len(word) - 1)
                            old_sym_new_action = self.__make_old_symbol_new_plam_action(action.name)
                            #~ old_sym_new_action = pair_action
                            new_remaining[new_non_terminal] = [(word[1:], new_sym_action)]
                            new_dict[non_terminal] += [([first, new_non_terminal], old_sym_new_action)]
                            new_actions_list += [new_sym_action, old_sym_new_action]
                            #~ print "        added new non_terminal:",new_non_terminal
            remaining = new_remaining
            #~ print "END PASS ************************************************************************"
        return new_dict, new_actions_list
    
    def __copy_composite_rule_table_with_actions(self, rules):
        rules_copy = {}
        for non_terminal in rules:
            alternatives = rules[non_terminal]
            new_alternatives = []
            for word, action in alternatives:
                new_alternatives += [(copy.deepcopy(word), action)]
            rules_copy[non_terminal] = new_alternatives
        return rules_copy
    
    def __save_actions_module(self, 
                                package_path, 
                                module_name, 
                                modules_to_import=None, 
                                additional_actions=None, 
                                verbose=False):
        if additional_actions is None:
            additional_actions = []
        lines = ['\n','# CFG actions\n\n']
        for module_to_import in modules_to_import:
            lines += ["from " + module_to_import + " import *\n"]
        lines += ['\n']
        lines += ['# Generated rule actions\n\n']
        action_names = self.actions_table.keys()
        action_names.sort()
        lines += [str(self.actions_table[action]) + '\n' 
                    for action in action_names
                    if action not in self.original_actions
                    and action is not '__builtins__']
        lines += ['\n']
        if additional_actions is not None and len(additional_actions) > 0:
            lines += ['# Additional actions\n\n']
            lines += [str(action) + '\n' for action in additional_actions]
            lines += ['\n']

        path = os.path.join(package_path, module_name)
        current = os.getcwd()
        os.chdir(package_path)
        the_file = file(module_name + '.py', 'w')
        the_file.writelines(lines)
        the_file.close()
        os.chdir(current)
        if verbose:
            print_time("* Saved actions in file '" + path + "'")
        
    def __save_dict_repr_module(self, package_path, module_name, verbose=False):
        text = "\n"
        text += "import aperiot.lexer\n\n"
        text += self.dict_repr_str()
        path = os.path.join(package_path, module_name)
        current = os.getcwd()
        os.chdir(package_path)
        the_file = file(module_name + '.py', 'w')
        the_file.write(text)
        the_file.close()
        os.chdir(current)
        if verbose:
            print_time("* Saved python representation in file '" + path + "'")
    
    # Private mutators
    
    def __collect_actions_garbage(self):
        #~ print "\ncollecting actions garbage"
        initial_size = len(self.actions_table)
        #~ print "  current actions:",str(initial_size)
        self.__count_action_references()
        marked_for_elimination = [action for action in self.actions_table if self.actions_table[action].ocurrences == 0]
        #~ print "  marked_for_elimination:",str(marked_for_elimination)
        for action_name in marked_for_elimination:
            action = self.actions_table[action_name]
            self.__mark_callees(action)
        marked_for_elimination = [action for action in self.actions_table if self.actions_table[action].ocurrences == 0]
        #~ print "  all marked_for_elimination:",str(marked_for_elimination)
        for action_name in marked_for_elimination:
            del self.actions_table[action_name]
        final_size = len(self.actions_table)
        #~ print "  actions after cleanup:",str(final_size)
        #~ print "  actions removed:",str(initial_size-final_size)

    def __count_action_references(self):
        for action_name in self.actions_table:
            action = self.actions_table[action_name]
            action.ocurrences = 0
            action.crosslinks = []
        for action_name in self.actions_table:
            action = self.actions_table[action_name]
            for var in action.term.freevars():
                self.actions_table[var].ocurrences += 1
                self.actions_table[var].crosslinks += [action_name]
        i = 0
        for rule in self.rules:
            i += 1
            action = rule.action
            action.ocurrences += 1
            action.crosslinks += [(i,str(rule))]
        
    def __generate_flat_rule_table(self, composite_rule_table):
        rules = composite_rule_table
        self.non_terminals = SymbolSet([NonTerminal(key) for key in rules])
        self.terminals = SymbolSet([])
        rule_list = []
        for item in composite_rule_table:
            for rhs, action in composite_rule_table[item]:
                new_rhs = []
                for symbol in rhs:
                    if isinstance(symbol, Token) or type(symbol) is ClassType and issubclass(symbol, Token):
                        new_terminal = Terminal(symbol)
                        new_rhs.append(new_terminal)
                        self.terminals.add(new_terminal)
                    else:
                        new_non_terminal = NonTerminal(symbol)
                        new_rhs.append(new_non_terminal)
                if len(new_rhs) == 0:
                    new_rhs.append(EmptyString())
                rule_list.append(Rule(NonTerminal(item), new_rhs, action))
        self.rules = rule_list
    
    def __generate_plamterms_from_action_templates(self, composite_rule_table):
        new_dict = {}
        for item in composite_rule_table:
            new_alternatives = []
            alt_counter = 0
            for rhs, template in composite_rule_table[item]:
                alt_counter += 1
                name = item.lower() + '_action_' + str(alt_counter)
                new_action = PlamRuleAction.make_from_template(name, rhs, template)
                new_alternatives += [(rhs, new_action)]
            new_dict[item] = new_alternatives
        return new_dict
    
    def __generate_pyfuncs_from_action_templates(self, composite_rule_table):
        new_dict = {}
        for item in composite_rule_table:
            new_alternatives = []
            alt_counter = 0
            for rhs, template in composite_rule_table[item]:
                alt_counter += 1
                name = item.lower() + '_action_' + str(alt_counter)
                new_action = PyRuleAction.make_from_template(name, rhs, template)
                new_alternatives += [(rhs, new_action)]
            new_dict[item] = new_alternatives
        return new_dict
    
    def __mark_callees(self, action):
        for callee_name in action.term.freevars():
            callee = self.actions_table[callee_name]
            callee.ocurrences -= 1
            if callee.ocurrences == 0:
                self.__mark_callees(callee)

    def __populate_actions_table(self):
        associations = [(rule.action.name, rule.action) for rule in self.rules]
        self.actions_table = dict(associations)
    
    def __simplify_actions(self):
        #~ print "-- simplifying actions ------------------------------------------------------------------------"
        #~ print "\n-- actions table before simplifying \n"
        #~ print_dict(self.actions_table, value_width=200, value_delim='\n', items='v')
        for action_name in self.actions_table:
            #~ print "now simplifying '"+action_name+"'"
            action = self.actions_table[action_name]
            #~ print "  before:"
            #~ print indent(str(action),4)
            #~ print
            #~ action.body = action.body.reduce(self.actions_table)
            #~ action.term = Abs(action.args, action.body)
            #~ action.term = action.term.normalize(self.actions_table,6)
            action.body = action.body.normalize(self.actions_table, 6)
            action.term = Abs(action.args, action.body)
            #~ print "  after:"
            #~ print indent(str(action),4)
            #~ print
        #~ print "\n-- actions table after simplifying \n"
        #~ print_dict(self.actions_table, value_width=200, value_delim='\n', items='v')
        #~ print "-- actions simplified  ------------------------------------------------------------------------"
        
    
    """
    Private static methods
    """
    
    def __make_new_expansion_action(self, old_action_1, old_action_2, matches):
        #~ print 'a\n'+str(old_action_1)
        #~ print 'b\n'+str(old_action_2)
        #~ print 'matches = ',str(matches)
        #~ new_action_name = 'new_' + old_action_1.name
        if hasattr(self, 'efac'):
            self.efac += 1
        else:
            self.efac = 0
        new_action_name = 'expansion_' + str(self.efac)
        #~ new_action_name = self.ind_gen.get_new_name(base='X')
        ocurrences = len(matches)
        total_args = len(old_action_1.args) - ocurrences + ocurrences * len(old_action_2.args)
        #~ print 'ocurrences=',ocurrences
        #~ print 'total_args=',total_args
        args = ['x' + str(i) for i in range(total_args)]
        passed = []
        j = 0
        k = 0
        for i in range(len(old_action_1.args)):
            #~ print 'i = ',str(i),'j = ',str(j),'k = ',str(k)
            if k < len(matches) and i == matches[k]:
                sub_args = ['x' + str(l) for l in range(j, j + len(old_action_2.args))]
                sub_pass = ', '.join(sub_args)
                passed += [old_action_2.name + "(" + sub_pass + ")"]
                j += len(old_action_2.args)
                k += 1
            else:
                passed += ['x' + str(j)]
                j += 1
        passed = ', '.join(passed)
        body = "return " + old_action_1.name + "(" + passed + ")\n"
        new_action = PyRuleAction(new_action_name, args, body)
        #~ print 'resulting action:'
        #~ print str(new_action)
        return new_action
        
    @staticmethod
    def __make_new_symbol_action(new_non_terminal, old_action_name, n):
        new_symbol_action_name = new_non_terminal.lower()
        #~ print 'making new symbol action:',new_symbol_action_name
        #~ print_env(self.actions_dict)
        #~ new_symbol_action_name = self.name_gen.get_new_name(upper=True)
        #~ args = ['*args']
        #~ body = "lambda x: " + old_action_name + "(x, *args)\n"
        args = ['x' + str(i) for i in range(n)]
        passed = ', '.join(['x'] + args)
        body = "return lambda x: " + old_action_name + "(" + passed + ")\n"
        #~ new_symbol_action = RuleAction(new_symbol_action_name, args, body, environment=self.actions_dict)
        new_symbol_action = PyRuleAction(new_symbol_action_name, args, body)
        return new_symbol_action
    
    @staticmethod
    def __make_old_symbol_new_action(old_action_name):
        old_symbol_new_action_name = 'new_'+old_action_name.lower()
        #~ print 'making old symbol new action:',old_symbol_new_action_name
        #~ print_env(self.actions_dict)
        #~ old_symbol_new_action_name = self.name_gen.get_new_name(upper=True)
        args = ['first', 'continuation']
        body = "return continuation(first)\n"
        #~ old_symbol_new_action_def = RuleAction(old_symbol_new_action_name, args, body, environment=self.actions_dict)
        old_symbol_new_action_def = PyRuleAction(old_symbol_new_action_name, args, body)
        return old_symbol_new_action_def
    
    @staticmethod
    def __make_pair_action():
        old_symbol_new_action_name = 'pair_action'
        #~ print 'making old symbol new action:',old_symbol_new_action_name
        #~ print_env(self.actions_dict)
        #~ old_symbol_new_action_name = self.name_gen.get_new_name(upper=True)
        args = ['first', 'continuation']
        body = "return continuation(first)\n"
        #~ old_symbol_new_action_def = RuleAction(old_symbol_new_action_name, args, body, environment=self.actions_dict)
        old_symbol_new_action_def = PyRuleAction(old_symbol_new_action_name, args, body)
        return old_symbol_new_action_def
    
    def __make_new_expansion_plam_action(self, old_action_1, old_action_2, matches):
        #~ print 'a\n'+str(old_action_1)
        #~ print 'b\n'+str(old_action_2)
        #~ print 'matches = ',str(matches)
        new_action_name = 'new_' + old_action_1.name
        if hasattr(self, 'efac'):
            self.efac += 1
        else:
            self.efac = 0
        new_action_name = 'expansion_' + str(self.efac)
        #~ new_action_name = self.ind_gen.get_new_name(base='X')
        ocurrences = len(matches)
        total_args = len(old_action_1.args) - ocurrences + ocurrences * len(old_action_2.args)
        #~ print 'ocurrences=',ocurrences
        #~ print 'total_args=',total_args
        args = ['x' + str(i) for i in range(total_args)]
        passed = []
        j = 0
        k = 0
        for i in range(len(old_action_1.args)):
            #~ print 'i = ',str(i),'j = ',str(j),'k = ',str(k)
            if k < len(matches) and i == matches[k]:
                sub_args = [Var('x' + str(l)) for l in range(j, j + len(old_action_2.args))]
                passed += [App(Var(old_action_2.name), sub_args)]
                j += len(old_action_2.args)
                k += 1
            else:
                passed += [Var('x' + str(j))]
                j += 1
        body = App(Var(old_action_1.name), passed)
        #~ print 'exp: ',str(body.__class__.__name__),' | \n  ',str(body)
        #~ print_dict(self.environment)
        #~ body = body.reduce(self.environment, 2)
        #~ print 'red:'
        #~ print '  ',str(body.__class__.__name__),' | ',str(body)
        #~ if not isinstance(body, LambdaTerm):
            #~ body = Constant(body)
        #~ print 'exp: ',str(body.__class__.__name__),' | ',str(body)
        new_action = PlamRuleAction(new_action_name, args, body)
        #~ print 'resulting action:'
        #~ print str(new_action)
        return new_action
        
    @staticmethod
    def __make_new_symbol_plam_action(new_non_terminal, old_action_name, n):
        new_symbol_action_name = new_non_terminal.lower()
        #~ print 'making new symbol action:',new_symbol_action_name
        #~ print_env(self.actions_dict)
        #~ new_symbol_action_name = self.name_gen.get_new_name(upper=True)
        #~ args = ['*args']
        #~ body = "lambda x: " + old_action_name + "(x, *args)\n"
        args = [Var('x' + str(i)) for i in range(n)]
        passed = [Var('_yyy')] + args
        body = Abs(['_yyy'], App(Var(old_action_name), passed))
        new_symbol_action = PlamRuleAction(new_symbol_action_name, args, body)
        return new_symbol_action
    
    @staticmethod
    def __make_old_symbol_new_plam_action(old_action_name):
        old_symbol_new_action_name = 'new_'+old_action_name.lower()
        #~ print 'making old symbol new action:',old_symbol_new_action_name
        #~ print_env(self.actions_dict)
        #~ old_symbol_new_action_name = self.name_gen.get_new_name(upper=True)
        args = ['first', 'continuation']
        body = App(Var('continuation'), [Var('first')])
        old_symbol_new_action_def = PlamRuleAction(old_symbol_new_action_name, args, body)
        return old_symbol_new_action_def
    
    @staticmethod
    def __make_pair_plam_action():
        old_symbol_new_action_name = 'pair_action'
        #~ print 'making old symbol new action:',old_symbol_new_action_name
        #~ print_env(self.actions_dict)
        #~ old_symbol_new_action_name = self.name_gen.get_new_name(upper=True)
        args = ['first', 'continuation']
        body = App(Var('continuation'), [Var('first')])
        old_symbol_new_action_def = PlamRuleAction(old_symbol_new_action_name, args, body)
        return old_symbol_new_action_def
    
    """
    Maybe methods
    """
    def __import_modules(self, package_name=None):
        if package_name is None:
            package_name = self.package_name
        #~ if package_name is None or package_name == '':
            #~ print "No package name given. No modules to import."
            #~ return
        original_cfg_module_name     = package_name + '.' + package_name + '_cfg_wan'
        original_actions_module_name = package_name + '.' + package_name + '_actions'
        ll1_cfg_module_name          = package_name + '.' + package_name + '_ll1_cfg_wan'
        ll1_actions_module_name      = package_name + '.' + package_name + '_ll1_actions'
        self.original_cfg_module     = CFG.__import_module(original_cfg_module_name)
        self.original_actions_module = CFG.__import_module(original_actions_module_name)
        self.ll1_cfg_module          = CFG.__import_module(ll1_cfg_module_name)
        self.ll1_actions_module      = CFG.__import_module(ll1_actions_module_name)
        self.rule_actions_module     = self.ll1_actions_module
        self.all_modules = \
            {
                'original_cfg':     self.original_cfg_module,
                'original_actions': self.original_actions_module,
                'll1_cfg':          self.ll1_cfg_module,
                'll1_actions':      self.ll1_actions_module,
                'rule_actions':     self.rule_actions_module
            }
        self.all_names = []
        for key in self.all_modules:
            self.all_names += dir(self.all_modules[key])

    def __incorporate_actions(self):
        for object_name in self.original_actions_module.__dict__:
            print "incorporating original action '"+object_name+"'"
            object = self.original_actions_module.__dict__[object_name]
            if type(object) is FunctionType:
                print "...it is a function"
                self.actions_table[object_name] = RuleAction.make_from_pyfunc(object)
        self.original_actions = copy.deepcopy(self.actions_table)
        for object_name in self.ll1_actions_module.__dict__:
            print "incorporating ll1 action '"+object_name+"'"
            object = self.ll1_actions_module.__dict__[object_name]
            if type(object) is FunctionType:
                print "...it is a function"
                self.actions_table[object_name] = RuleAction.make_from_pyfunc(object)
                
    
    @staticmethod
    def __import_module(full_module_name):
        try:
            exec "import " + full_module_name
            module = eval(full_module_name)
            bytecode_file = module.__file__ + 'c' 
            if os.path.exists(bytecode_file):
                os.remove(bytecode_file)
            reload(module)
            return module
        except ImportError:
            import aperiot.empty
            return aperiot.empty
    
    @staticmethod
    def __make_new_empty_word_action(non_terminal):
        args = []
        body = "lambda x: x\n"
        return RuleAction(non_terminal.lower(), args, body)
    
    """
    Old methods
    """

    def __generate_from_dict(self, rules):
        self.non_terminals = SymbolSet([NonTerminal(key) for key in rules])
        self.terminals = SymbolSet([])
        new_dict = {}
        rule_list = []
        for item in rules:
            new_alts = []
            alt_counter = 0
            for alt in rules[item]:
                alt_counter += 1
                new_action_name = item.lower() + '_action_' + str(alt_counter)
                if type(alt) is tuple:
                    rhs, action = alt
                elif type(alt) is list:
                    rhs, action = alt, new_action_name
                if type(action) is str:
                    if action in self.actions_table:
                        print action,'is in the dict'
                        action = self.actions_table[action]
                    else:
                        print action,'is assumed to be a template'
                        action = RuleAction.make_from_template(new_action_name, rhs, action, self.actions_table)
                elif type(action) is FunctionType:
                    print action.__name__,'is a function'
                    action = RuleAction.make_from_pyfunc(action, self.actions_table)
                print 'action: ',action.__name__
                print 'type:',type(action)
                self.actions_table[action.name] = action
                new_alts += [(rhs, action)]
                new_rhs = []
                for symbol in rhs:
                    if isinstance(symbol, Token) or type(symbol) is ClassType and issubclass(symbol, Token):
                        new_terminal = Terminal(symbol)
                        new_rhs.append(new_terminal)
                        self.terminals.add(new_terminal)
                    else:
                        new_non_terminal = NonTerminal(symbol)
                        new_rhs.append(new_non_terminal)
                if len(new_rhs) == 0:
                    new_rhs.append(EmptyString())
                rule_list.append(Rule(NonTerminal(item), new_rhs, action))
            new_dict[item] = new_alts
        self.rules = rule_list
        self.dict_repr = new_dict
    


class Rule:
    
    def __init__(self, lhs, rhs, action=None):
        if not isinstance(lhs, NonTerminal):
            raise CFGException('LHS must be non-terminal')
        for item in rhs:
            if not isinstance(item, Terminal) and not isinstance(item, NonTerminal) and not isinstance(item, EmptyString):
                raise CFGException('RHS can contain only terminals, non-terminals, or the empty string')
            if isinstance(item, EmptyString) and len(rhs) > 1:
                raise CFGException('RHS contains the empty string and more than one other symbol. If the RHS contains more the empty string it cannot contain any other symbol.')
        self.lhs = lhs
        self.rhs = rhs
        self.action = action
        
    def __str__(self):
        rhs_str_list = [str(sym) for sym in self.rhs]
        rhs_str = reduce(lambda a, b: a + ' ' + b, rhs_str_list, "")
        action_name = self.action
        if callable(self.action):
            action_name = self.action.__name__
        return str(self.lhs) + " -> " \
            +  rhs_str \
            +  " : " + action_name




class RuleAction:
    
    def __init__(self, name, args=None, body=None, environment=None, rule=None):
        self.name = name
        self.__name__ = name
        if type(args) is list and all(type(item) is str for item in args):
            self.args = args
        elif type(args) is int:
            self.args = ['x' + str(i) for i in range(args)]
        else:
            self.args = [str(item) for item in args]
        self.body = body
        self.environment = {}
        if environment is not None:
            self.environment = dict(environment)
        self.environment[self.name] = self
        self.rule = rule


class PlamRuleAction(RuleAction):
    
    def __init__(self, name, args=None, body=None, environment=None, rule=None):
        RuleAction.__init__(self, name, args, body, environment=environment, rule=rule)
        self.term = Abs(self.args, self.body)
        self.ocurrences = 0
        self.crosslinks = []
    
    def __call__(self, *args):
        #~ print "-- applying:"
        #~ print indent(str(self))
        #~ print "with args: "+str(args)
        #~ composition = App(self.term, args)
        temp_env = {}
        exec str(self) in self.environment
        #~ print "env: "+str(type(self.environment))
        #~ print_dict(self.environment)
        #~ outcome = composition.execute(self.environment)
        func = eval(self.name, self.environment, self.environment)
        outcome = func(*args)
        #~ print 'outcome type:',str(type(outcome))
        #~ print 'outcome class:',str(outcome.__class__.__name__)
        #~ print 'outcome:',str(outcome)
        return outcome

    def __copy__(self):
        return PlamRuleAction(self.name, self.args, self.body, self.environment, self.rule)

    def __getstate__(self):
        #~ print "* saving plam rule action: "+self.name
        #~ print_dict(self.__dict__)
        obj_dict = copy.copy(self.__dict__)
        del obj_dict['environment']
        if hasattr(self, 'function'):
            del obj_dict['function']
        return obj_dict
    
    def __setstate__(self, dict):
        self.__dict__.update(dict)
    
    def __str__(self):
        params = ', '.join(self.args)
        header = "def " + self.name + "(" + params + "):\n"
        body   = "  return " + str(self.body) + "\n"
        source = header + body
        return source
        
    @staticmethod
    def make_from_template(name, rhs, template, environment=None):
        #~ print 'making from template:',str(template)
        args = ['x' + str(i) for i in range(len(rhs))]
        template = string.Template(template.replace('$', '$arg'))
        subst = dict(('arg' + str(i + 1), args[i]) for i in range(len(args)))
        outcome = template.substitute(subst)
        body = Const(outcome)
        return PlamRuleAction(name, args, body, environment)


class PyRuleAction(RuleAction):
    
    def __init__(self, name, args=None, body='', environment=None, pyfunc=None, rule=None):
        RuleAction.__init__(self, name, args, body, environment=environment, rule=rule)
        if body == '':
            self.body ='return \n'
        args_str = ', '.join(self.args)
        self.header = "def " + name + "(" + args_str + "):\n"
        self.source = self.header + indent(self.body) + '\n'
        if pyfunc is not None and type(pyfunc) is FunctionType:
            self.function = pyfunc
        else:
            exec self.source in self.environment
            self.function = self.environment[self.name]
    
    @staticmethod
    def make_from_pyfunc(function, environment=None):
        #~ print 'making from pyfunc',function.__name__
        if type(function) is FunctionType:
            #~ print 'alt 1'
            name = function.__name__
            argspec = inspect.getargspec(function)
            args = argspec[0]
            try:
                source = inspect.getsource(function)
                body = '\n'.join((source.split('\n'))[1:])
                body = dedent(body, 4)
            except IOError:
                body = 'pass \n'
        elif type(function) is str:
            #~ print 'alt 2'
            source = function
            lines = source.split('\n')
            header = lines[0]
            body = '\n'.join(lines[1:])
            body = dedent(body, 4)
            sp_pos = header.index(' ')
            lpar_pos = header.index('(')
            rpar_pos = header.index(')')
            name = header[sp_pos + 1: lpar_pos].strip()
            args = [var for var in header[lpar_pos + 1: rpar_pos] if var is not '']
        #~ print '  args:',args
        #~ print '  body:'
        #~ print indent(body,4)
        return PyRuleAction(name, args, body, environment, function)
    
    @staticmethod
    def make_from_template(name, rhs, template, environment=None):
        #~ print 'making from template:',str(template)
        args = ['x' + str(i) for i in range(1, len(rhs) + 1)]
        template = string.Template(template.replace('$', '$arg'))
        subst = dict(('arg' + str(i + 1), args[i]) for i in range(len(args)))
        outcome = template.substitute(subst)
        body = "return " + outcome + "\n"
        return PyRuleAction(name, args, body, environment)

    def __call__(self, *args):
        #~ print '\ntrying to call function:\n'
        #~ print self.source
        #~ print '\nwith args: ('+str(args)+')'
        #~ print '\nenvironment: \n'
        #~ print_dict(self.environment)
        outcome = self.function(*args)
        #~ print 'result = ',str(outcome)
        return outcome
    
    def __copy__(self):
        return PyRuleAction(self.name, self.args, self.body, self.environment, self.function, self.rule)
    
    def __getstate__(self):
        #~ print "* saving py rule action: "+self.name
        #~ print_dict(self.__dict__)
        obj_dict = copy.copy(self.__dict__)
        del obj_dict['function'] 
        del obj_dict['environment']
        return obj_dict
    
    def __setstate__(self, dict):
        self.__dict__.update(dict)
    
    def __str__(self):
        return self.source

class Symbol:
    
    def __init__(self, sym):
        self.sym = sym
        
    def value(self):
        return self.sym
        
    def __eq__(self, other):
        if isinstance(other, Symbol):
            #~ print '+++:',self.sym,other.sym
            result = self.sym == other.sym
            #~ print '    ',str(result)
            return result
        #~ elif isinstance(other, Token):
            #~ return other == self.sym 
        return False
    
    def matches(self, token):
        return token == self.sym or self.sym.matches(token)
        
    def __ne__(self, other):
        return self.sym != other.sym
    
    def __hash__(self):
        return hash(self.sym)
        
    def __call__(self):
        return self.sym
        
    def __str__(self):
        string = str(self.sym)
        if type(self.sym) is ClassType:
            string = short_class_name(self.sym)
        return self.__class__.__name__ + '(' + string + ')'


class EmptyString(Symbol):
    symbol = Token('<the empty string>')
    def __init__(self):
        Symbol.__init__(self, EmptyString.symbol)
    def __eq__(self, other):
        return isinstance(other, EmptyString)
    def __str__(self):
        return 'EmptyStringSymbol'
    
 
class Terminal(Symbol):
    pass
    

class NonTerminal(Symbol):
    pass
    


class SymbolSet:
    
    def __init__(self, iterator):
        self.the_set = set(iterator)
    
    def __iter__(self):
        return iter(self.the_set)
    
    def __contains__(self, item):
        for element in self.the_set:
            if type(item) is str:
                return element == item
            elif isinstance(item, EmptyString):
                return isinstance(element, EmptyString)
            elif type(item) is InstanceType and isinstance(item, Token):
                return element.sym == item
            elif type(item) is ClassType and issubclass(item, Token):
                return element.sym.__class__ == item
            elif element.matches(item):
                return True
        return False
    
    def add(self, item):
        self.the_set.add(item)
    
    def union(self, other):
        return SymbolSet(self.the_set | other.the_set)
    
    def __len__(self):
        return len(self.the_set)
    
    def __str__(self):
        return str([str(item) for item in self.the_set])
        
        

class CFGException(Exception):
    pass




