"""
Module: lexer
Author: Ernesto Posse
Created on: April 12, 2006
Last Update: Jan 22, 2007

Description:
  Lexer for basic type expressions for the dyntypchk package.
"""

from posse_tools.tries import *
from posse_tools.stacks import *
from types import *
import copy

# Public classes

class Lexer:
    
    def __init__(self, 
                    tokens=[], 
                    string_indicator='"', 
                    comment_indicator='#', 
                    tab_size=4,
                    yield_newlines=True,
                    yield_indents=True):
        self.tokens = tokens
        self.string_indicator = string_indicator
        self.comment_indicator = comment_indicator
        self.tab_size = tab_size
        self.yield_newlines = yield_newlines
        self.yield_indents = yield_indents

        self.trie = Trie()
        for token in tokens:
            if isinstance(token, Token):
                self.trie[token.string] = token
            elif type(token) is str:
                self.trie[token] = Unknown(token)
            else:
                raise LexicographicException("The object " + repr(token) + " is not a Token.")
        
        self.readers = [self.__try_alphanumeric, 
                        self.__try_symbol, 
                        self.__try_numeric, 
                        self.__try_string, 
                        self.__try_comment, 
                        self.__giveup]
        self.pos = 0
        
    def __getstate__(self):
        #~ return {'dummy_lexer':'dummy'}
        obj_dict = copy.copy(self.__dict__)
        del obj_dict['readers']
        return obj_dict
    
    def __setstate__(self, dict):
        self.__dict__.update(dict)
        self.readers = [self.__try_alphanumeric, 
                        self.__try_symbol, 
                        self.__try_numeric, 
                        self.__try_string, 
                        self.__try_comment, 
                        self.__giveup]

    def original_tokenize(self, string):
        self.string = string
        self.pos = 0
        while self.pos < len(string):
            self.__skip_spaces()
            for reader in self.readers:
                success, token = reader()
                if success:
                    yield token
                    break
    
    def tokenize(self, source, yield_newlines=None, yield_indents=None):
        if yield_newlines is None:
            yield_newlines = self.yield_newlines
        if yield_indents is None:
            yield_indents = self.yield_indents
        if type(source) is str:
            source = StringLinesIterator(source)
        self.line = 0
        self.stack = Stack([0])
        self.line_counter = 0
        for line in source:
            self.line_counter += 1
            if line.endswith('\r\n'):
                line = line[:-2] + '\n'
            self.string = line
            self.pos = 0
            # Skip leading spaces
            self.__skip_spaces()
            # self.pos now has the indentation level
            # Ignore comments if any
            is_comment, token = self.__try_comment()
            if not is_comment and self.pos < len(self.string):  # if it is a blank line, ignore it
                # generate indent or dedent tokens according to the indentation level and the stack
                if yield_indents:
                    last_indent = self.stack.top()
                    if self.pos > last_indent:
                        self.stack.push(self.pos)
                        yield Indent(linenum=self.line_counter, col=self.pos)
                    elif self.pos < last_indent:
                        while self.stack.top() > self.pos:
                            self.stack.pop()
                            yield Dedent(linenum=self.line_counter, col=self.pos)
                        if self.stack.top() != self.pos:
                            raise LexicographicException('Too many or too few dedents at line '+str(self.line))
                # process the rest of the line
                comment = False
                empty_line = True
                while self.pos < len(self.string):
                    for reader in self.readers:
                        success, token = reader()
                        comment = isinstance(token, Comment)
                        if comment:
                            break
                        elif success:
                            empty_line = False
                            yield token
                            break
                    self.__skip_spaces()
                if yield_newlines and not comment:
                    last = len(line) - 1
                    if last >= 1 and line[last-1] == '\r':
                        last = last - 1
                    yield Newline(linenum=self.line_counter, col=last) 
            self.line += 1
        # generate remaining dedents
        if yield_indents:
            while self.stack.top() != 0: 
                self.stack.pop()
                yield Dedent(linenum=self.line_counter, col=self.pos)
    
    def __getitem__(self, string):
        return self.trie[string]
    
    def __setitem__(self, string, token):
        self.trie[string] = token
    
    def __skip_spaces(self):
        while self.pos < len(self.string) and self.string[self.pos].isspace():
            if self.string[self.pos] == '\t':
                self.pos += self.tab_size
            else:
                self.pos += 1
    
    def __try_alphanumeric(self):
        self.trie.reset()
        index = self.pos
        if index < len(self.string) and self.string[index].isalpha():
            start = self.pos
            identifier = ''
            tokens_found = []
            try:
                while index < len(self.string) and self.isidchar(self.string[index]):
                    data = self.trie.step(self.string[index])
                    if data is not None:
                        tokens_found.append(data)
                    identifier += self.string[index]
                    index += 1
            except KeyError:
                # If there still are more id chars, then the token is an identifier
                while index < len(self.string) and self.isidchar(self.string[index]):
                    identifier += self.string[index]
                    index += 1
                self.pos = index
                return (True, Identifier(identifier, 
                                            linenum=self.line_counter, 
                                            start_col=start, end_col=index - 1))
            self.pos = index
            if len(tokens_found) == 0:
                return (True, Identifier(identifier, 
                                            linenum=self.line_counter, 
                                            start_col=start, end_col=index - 1))
                #~ raise LexicographicException("There is no symbol associated with this token: '" \
                                              #~ + identifier + "'")
            token = copy.copy(tokens_found[len(tokens_found)-1])
            token.set_pos(self.line_counter, start, index - 1)
            return (True, token)
        return (False, None)
    
    def __try_symbol(self):
        self.trie.reset()
        index = self.pos
        if index < len(self.string) and self.issymbol(self.string[index]):
            start = self.pos
            symbol = ''
            tokens_found = []
            try:
                while index < len(self.string) and self.issymbol(self.string[index]):
                    data = self.trie.step(self.string[index])
                    if data is not None:
                        tokens_found.append(data)
                    symbol += self.string[index]
                    index += 1
            except KeyError:
                if len(tokens_found) == 0:
                    return (False, Unknown(symbol, 
                                            linenum=self.line_counter, 
                                            start_col=start, end_col=index - 1))
            self.pos = index
            if len(tokens_found) == 0:
                raise LexicographicException("There is no symbol associated with this token: '" \
                                              + symbol + "'")
            token = copy.copy(tokens_found[len(tokens_found)-1])
            token.set_pos(self.line_counter, start, index - 1)
            return (True, token)
        return (False, None)
        
    def __try_numeric(self):
        index = self.pos
        if index < len(self.string) and self.string[index].isdigit():
            start = self.pos
            literal = ''
            while index < len(self.string) and self.string[index].isdigit():
                literal += self.string[index]
                index += 1
            if index < len(self.string) and self.string[index] == '.':
                literal += self.string[index]
                index += 1
                while index < len(self.string) and self.string[index].isdigit():
                    literal += self.string[index]
                    index += 1
            self.pos = index
            return (True, NumberLiteral(literal, 
                                        linenum=self.line_counter, 
                                        start_col=start, end_col=index - 1))
        return (False, None)
    
    def __try_string(self):
        index = self.pos
        if index < len(self.string) and self.string[index] == self.string_indicator:
            start = self.pos
            literal = ''
            index += 1
            while index < len(self.string) and self.string[index] != self.string_indicator:
                literal += self.string[index]
                index += 1
            if index >= len(self.string):
                raise LexicographicException('Run-off string at position ' + str(self.pos))
            self.pos = index + 1
            return (True, StringLiteral(literal, 
                                        linenum=self.line_counter,
                                        start_col=start, end_col=index))
        return (False, None)
    
    def __try_comment(self):
        index = self.pos
        if index < len(self.string) and self.string[index] == self.comment_indicator:
            comment = self.string[index]
            index += 1
            while index < len(self.string) and self.string[index] != '\n':
                comment += self.string[index]
                index += 1
            self.pos = index
            return (True, Comment(comment))
        return (False, None)
    
    def __giveup(self):
        index = self.pos
        self.pos += 1
        return (True, Unknown(self.string[index], 
                                linenum=self.line_counter,
                                start_col=index, end_col=index))

    def isidchar(self, char):
        return char.isalnum() or char == '_'
        
    def issymbol(self, char):
        return not char.isspace() and not char.isalnum() and char != '_' \
            and char != self.string_indicator and char != self.comment_indicator


class Token:
    
    def __init__(self, string, symbolic_name=None, linenum=None, start_col=None, end_col=None):
        self.string = str(string)
        self.value = self.string
        self.linenum = linenum
        self.start_col = start_col
        self.end_col = end_col
        self.pos = (linenum, start_col, end_col)
        #~ if type(start_col) is int:
            #~ self.start_col = start_col + 1
        #~ if type(end_col) is int:
            #~ self.end_col = end_col + 1
        if symbolic_name is None:
            self.symbolic_name = str(string)
        else:
            self.symbolic_name = str(symbolic_name)
    
    def set_pos(self, line, start, end):
        self.linenum = line
        self.start_col = start
        self.end_col = end
        self.pos = (line, start, end)
    
    def __eq__(self, other):
        if isinstance(other, str):
            return self.string == other
        if type(other) is ClassType:
            return False
        if self.__class__ is not other.__class__:
            return False
        if isinstance(other, Token):
            return self.symbolic_name == other.symbolic_name
        return False
    
    def family(self):
        return self
    
    def matches(self, token):
        return self == token
    
    def __ne__(self, other):
        return self.string != other.string
    
    def __hash__(self):
        return hash(self.string)
    
    def __str__(self):
        optional = ""
        if self.symbolic_name is not None and self.symbolic_name != self.string:
            optional = "', '" + self.symbolic_name
        return self.__class__.__name__ + "('" + self.string + optional + "')"


class Keyword(Token):
    pass


class Identifier(Token):
    
    symbolic_name = 'identifier'
    
    def family(self):
        return self.__class__
        
    def matches(self, token):
        if type(token) is InstanceType and isinstance(token, Identifier):
            return True
        if type(token) is ClassType and issubclass(token, Identifier):
            return True
        return False

class NumberLiteral(Token):
    
    symbolic_name = 'number'
    
    def family(self):
        return self.__class__
        
    def matches(self, token):
        if type(token) is InstanceType and isinstance(token, NumberLiteral):
            return True
        if type(token) is ClassType and issubclass(token, NumberLiteral):
            return True
        return False
    def __int__(self):
        return int(self.string)
    def __float__(self):
        return float(self.string)
    def val(self):
        try:
            val = int(self.string)
            return val
        except ValueError:
            val = float(self.string)
            return val

class StringLiteral(Token):

    symbolic_name = 'string'
    
    def family(self):
        return self.__class__
        
    def matches(self, token):
        if type(token) is InstanceType and isinstance(token, StringLiteral):
            return True
        if type(token) is ClassType and issubclass(token, StringLiteral):
            return True
        return False

class Operator(Token):
    pass


class Separator(Token):
    pass


class Bracket(Token):
    pass


class Newline(Token):
    
    symbolic_name = 'newline'

    def __init__(self, linenum=None, col=None):
        Token.__init__(self, '', 'newline', linenum, col, col)
    
    def family(self):
        return self.__class__
        
    def matches(self, token):
        if type(token) is InstanceType and isinstance(token, Newline):
            return True
        if type(token) is ClassType and issubclass(token, Newline):
            return True
        return False


class Indent(Bracket):
    
    symbolic_name = 'indent'

    def __init__(self, linenum=None, col=None):
        Bracket.__init__(self, '', 'indent', linenum, col, col)
    
    def family(self):
        return self.__class__
        
    def matches(self, token):
        if type(token) is InstanceType and isinstance(token, Indent):
            return True
        if type(token) is ClassType and issubclass(token, Indent):
            return True
        return False

class Dedent(Bracket):
    
    symbolic_name = 'dedent'
    
    def __init__(self, linenum=None, col=None):
        Bracket.__init__(self, '', 'dedent', linenum, col, col)
    
    def family(self):
        return self.__class__
        
    def matches(self, token):
        if type(token) is InstanceType and isinstance(token, Dedent):
            return True
        if type(token) is ClassType and issubclass(token, Dedent):
            return True
        return False

class Comment(Token):
    pass

class Unknown(Token):
    
    symbolic_name = 'unknown'


class EndOfInput(Token):
    symbol = '$EndOfInput$'
    def __init__(self):
        Token.__init__(self, EndOfInput.symbol)
    #~ def __eq__(self, other):
        #~ return other.__class__ == self.__class__

class LexicographicException(Exception):
    pass

class StringLinesIterator:

    def __init__(self, string):
        self.string = string
    
    def __iter__(self):
        self.pos = 0
        return self
        
    def next(self):
        if self.pos >= len(self.string):
            raise StopIteration
        next_line = ''
        while self.pos < len(self.string) and self.string[self.pos] != '\n':
            next_line += self.string[self.pos]
            self.pos += 1
        self.pos += 1 
        next_line += '\n'
        return next_line

