"""
Module: llparser
Author: Ernesto Posse
Created on: April 13, 2006
Last update: Jan 22, 2007

Description:
  Generic LL(1) parser

Typical use:
    See file 'parspattern.py'

Classes:
    Parser
        Attributes:
            parse_tree: ParseTreeNode       The last tree computed
            cfg: CFG                        The Context Free Grammar of the parser
            table: ParsingTable             The parsing table corresponging to the CFG
        Methods:
            __init__(self: Parser, cfg: CFG, dicard_choices:bool=True)
                Constructor. Creates a new parser from the given CFG. 
                If discard_choices is True, the resulting parsing table is LL(1), that
                is, all its elements are single rule numbers. Otherwise the elements
                of the table are sets of applicable rule numbers. 
                (Note: The parser currently does not support parsing for LL(k) tables 
                with k>1)
            parse(self: Parser, 
                  input_stream: Token iterator, 
                  apply_actions:bool=True, 
                  pass_all:bool=True) -> 'a
                Parses the input stream of tokens into a concrete syntax tree. If the
                apply_actions parameter is True, it applies the actions associated with 
                rules in the CFG to the tree. The outcome of applying these actions is 
                returned. 
                    If pass_all is True, each action is passed as arguments all outcomes 
                associated with both terminal and non-terminal symbols in the RHS of the 
                rule. These arguments are passed in order from left-to-right as they 
                appear in the RHS of the rule. If pass_all is False, the action receives 
                as input only the outcomes associated with non-terminals in the RHS.
                    The input_stream can be any object that supports the iterator protocol.
                In particular it must have an __iter__ method which must return an
                iterator object, i.e. an object with a next method. Normally this 
                would be a list or a generator. 
                    The objects contained in the input stream must by instances of the 
                Token class. The outcome of the Lexer.tokenize method satisfies these 
                conditions.
            get_parse_tree(self: Parser, input_stream: Token iterator) -> ParseTreeNode
                This is the core algorithm for the LL(1) parser. The parameters are
                as in the parse method above. Its outcome is the root node of the
                Concrete Syntax Tree (CST).
            get_derivation(self: Parser, 
                           input_stream: Token iterator, 
                           pass_all:bool=True) -> int list
                This is like the get_parse_tree method, but instead of returning the CST,
                its outcome is the left-most derivation of the input_stream given as a 
                list of CFG rule numbers, appearing in the order in which the rules are 
                applied.
            build_parse_tree(self: Parser, 
                             rule_sequence: int list, 
                             include_terminals:bool=True) -> ParseTreeNode
                This method takes a list of CFG rule numbers (the rule_sequence) and
                returns a pointer to the root of the concrete syntax parse tree.
            apply_actions(self: Parser, node: ParseTreeNode, pass_all:bool=True) -> 'a
                This method applies recursively the actions associated with the rules
                to the concrete syntax tree given (starting from node, doing a 
                depth-first traversal) It returns the outcome of these actions.

    ParsingTable
    ParseTreeNode
    ParsingException
"""

from lexer import *
from cfgs import *
from posse_tools.stacks import *
from types import *
from copy import *


class Parser:
    
    def __init__(self, 
                    cfg, 
                    lexer=None, 
                    rule_actions_module=None, 
                    discard_choices=True,
                    verbose=False):
        self.table = ParsingTable(cfg, discard_choices, verbose)
        self.cfg = cfg
        self.lexer = lexer
        self.rule_actions_module = rule_actions_module
    
    """
    Public instance methods
    """
    
    # Public accessors, selectors and extractors
    
    def __getstate__(self):
        #~ return {'dummy_parser':'dummy'}
        obj_dict = copy(self.__dict__)
        if type(self.rule_actions_module) is ModuleType:
            obj_dict['rule_actions_module'] = self.rule_actions_module.__name__
        return obj_dict
    
    def parse(self, input_stream, apply_actions=True, pass_all=True):
        if type(input_stream) is str or type(input_stream) is file:
            input_stream = self.lexer.tokenize(input_stream)
        self.parse_tree = self.get_parse_tree(input_stream)
        if apply_actions:
            return self.apply_actions(self.parse_tree)
        return self.parse_tree
        
        #~ success, derivation = self.get_derivation(input_stream)
        #~ if success:
            #~ self.parse_tree = self.build_parse_tree(derivation)
            #~ outcome = self.apply_actions(self.parse_tree, pass_all)
            #~ return outcome
        
    def get_parse_tree(self, input_stream):
        success = False
        eoi_sym = Terminal(EndOfInput())
        input_stream = iter(input_stream)
        input_stream = add_eoi(input_stream)
        root = None
        node_counter = 0
        stack = Stack([(self.cfg.start, root), (eoi_sym, None)])
        i = 0
        current = input_stream.next()
        try:
            while True:
                top, parent_node = stack.top()
                #~ print "i = "+str(i)+" current = '"+str(current)+"' top = '"+str(top) + "'"
                #~ print_stack(stack)
                #~ if parent_node is not None:
                    #~ parent_node.pprint()
                if top == eoi_sym:
                    #~ print '--> top is eoi_sym'
                    if current == eoi_sym.value():
                        #~ print '--> current is eoi'
                        success = True
                        break
                    else:
                        #~ print '--> current is not eoi'
                        #~ (line,col) = position_in_text(input_stream, i)
                        raise ParsingException("I wasn't expecting anything else, but there " \
                                                + "still are some unread tokens, at line " + str(current.linenum) \
                                                + " in column " + str(current.start_col)+".", current)
                elif isinstance(top, Terminal):
                    if current.matches(top.value()):
                        #~ print "  match at: "+str(i)+" current = '"+str(current)+"'"
                        stack.pop()
                        new_node = ParseTreeNode(node_counter, Terminal(current))
                        node_counter += 1
                        if parent_node is not None: # and parent_node.symbol != Terminal(current):
                            #~ print '  parent = ' + str(parent_node.identifier)
                            parent_node.add_child(new_node)
                        elif parent_node is None:
                            #~ print '  new root = ' + str(new_node.identifier)
                            root = new_node
                        current = input_stream.next()
                        i += 1
                    else:
                        raise ParsingException("I was expecting '" + str(top.sym) \
                                                + "' but read '" + str(current) \
                                                + "' instead, at line " + str(current.linenum)\
                                                + " column " + str(current.start_col) + ".", \
                                                current)
                elif isinstance(top, NonTerminal):
                    if current == eoi_sym.value():
                        rule_number = self.table[top, EmptyString.symbol]
                    else:
                        try:
                            #~ current in self.cfg.terminals:
                            rule_number = self.table[top, current.family()]
                        except KeyError:
                            #~ print self.cfg.terminals
                            raise ParsingException("Illegal token: '" + str(current) \
                                                    + "' at line " + str(current.linenum) \
                                                    + " from column " + str(current.start_col) \
                                                    + " to column " + str(current.end_col)+".", current)
                    if rule_number is None:
                        firsts = '  '
                        full_first_list = list(self.table.non_terminal_first_sets[top])
                        if len(full_first_list) > 0:
                            firsts_list = full_first_list[:-1]
                            last = full_first_list[-1]
                            firsts += ', '.join([short_str(token.sym, 50) for token in firsts_list])
                            if len(firsts_list) > 0:
                                firsts += ' or ' + short_str(last.sym, 50)
                            else:
                                firsts += short_str(last.sym, 50)
                        firsts += "\n"
                        raise ParsingException("No rule to apply. "+"Illegal token: '" + str(current) \
                                                + "' at line " + str(current.linenum) \
                                                + " from column " + str(current.start_col) \
                                                + " to column "+str(current.end_col)+". " \
                                                + "Non-terminal symbol expected: "+str(top.sym)+". " \
                                                + "I was expecting one of the following tokens: \n" \
                                                + firsts, 
                                                current)
                    rule = self.cfg[rule_number]
                    #~ print "  using rule #"+str(rule_number)+"  Rule: "+str(rule)
                    new_node = ParseTreeNode(node_counter, top, rule_number)
                    node_counter += 1
                    if parent_node is not None:
                        #~ print '  parent = ' + str(parent_node.identifier)
                        parent_node.add_child(new_node)
                    else:
                        #~ print '  new root = ' + str(new_node.identifier)
                        root = new_node
                    stack.pop()
                    reverse_rhs = reversed(rule.rhs)
                    for symbol in reverse_rhs:
                        stack.push((symbol, new_node))
                elif isinstance(top, EmptyString):
                    #~ print '---- empty string'
                    stack.pop()
                    new_node = ParseTreeNode(node_counter, EmptyString())
                    node_counter += 1
                    if parent_node is not None: # and parent_node.symbol != Terminal(current):
                        #~ print '  parent = ' + str(parent_node.identifier)
                        parent_node.add_child(new_node)
                    elif parent_node is None:
                        #~ print '  new root = ' + str(new_node.identifier)
                        root = new_node
                else:
                    raise ParsingException("There is an invalid element in the grammar: '"+str(top)+"'")
        except StopIteration:
            pass
        #~ root.pprint()
        return root

    def get_derivation(self, input_stream):
        success = False
        eoi = EndOfInput()
        input_stream = iter(input_stream)
        input_stream = add_eoi(input_stream)
        stack = Stack([self.cfg.start, eoi])
        derivation = []
        i = 0
        current = input_stream.next()
        #~ while i < len(input_stream):
        #~ for current in input_stream:
        try:
            while True:
                #~ current = input_stream[i]
                top = stack.top()
                #~ print "i = "+str(i)+" current = '"+str(current)+"' top = '"+str(top)+"' stack = "+str(stack)
                if top == eoi:
                    if current == eoi:
                        success = True
                        break
                    else:
                        (line,col) = position_in_text(input_stream, i)
                        raise ParsingException("I wasn't expecting anything else, but there still are some unread tokens, at position: "+str(i)+" line#"+str(line)+", col#"+str(col))
                elif isinstance(top, Terminal):
                    if top.value() == current:
                        #~ print "  match at: "+str(i)+" current = '"+str(current)+"'"
                        stack.pop()
                        current = input_stream.next()
                        i += 1
                    else:
                        raise ParsingException("I was expecting '"+str(top)+"' but read '"+str(current)+"' instead.")
                elif isinstance(top, NonTerminal):
                    if current == eoi:
                        rule_number = self.table[top.value(), EmptyString.symbol]
                    elif current in self.cfg.terminals:
                        rule_number = self.table[top.value(), current]
                    else:
                        raise ParsingException("Illegal token: '" + str(current) + "' at position " + str(i) + "; context: " + str(current))
                    if rule_number is None:
                        raise ParsingException("No rule to apply")
                    rule = self.cfg[rule_number]
                    #~ print "  using rule #"+str(rule_number)+"  Rule: "+str(rule)
                    derivation.append(rule_number)
                    stack.pop()
                    reverse_rhs = reversed(rule.rhs)
                    for symbol in reverse_rhs:
                        stack.push(symbol)
                elif isinstance(top, EmptyString):
                    stack.pop()
                else:
                    raise ParsingException("There is an invalid element in the grammar: '"+str(top)+"'")
        except StopIteration:
            pass
        #~ print "parsing success = "+str(success)
        #~ print "derivation numbers = "+str(derivation)
        #~ print "derivation = "+str([str(self.cfg[i]) for i in derivation])
        return (success, derivation)

    def build_parse_tree(self, rule_sequence, include_terminals=True):
        #~ print "building parse tree"
        root = ParseTreeNode(0, self.cfg.start)
        counter = 0
        stack = Stack([(self.cfg.start, root)])
        for rule_id in rule_sequence:
            rule = self.cfg[rule_id]
            #~ print "processing rule #"+str(rule_id)+": "+str(rule)
            top_sym, top_node = stack.pop()
            #~ print "top = "+str(top_sym)+" stack = "+str(stack)
            #~ print "  "+top_sym.__class__.__name__
            while isinstance(top_sym, Terminal):
                top_sym, top_node = stack.pop()
                #~ print "top = "+str(top_sym)+" stack = "+str(stack)
            if top_sym != rule.lhs:
                raise ParsingException("Inconsistency: the stack's top and the rule's LHS do not match.")
            #node['children'] = [{'symbol':sym, 'children':[]} for sym in rule.rhs if isinstance(sym, NonTerminal)]
            children = []
            #~ reverse_rhs = [rule.rhs[j] for j in range(len(rule.rhs) - 1, -1, -1)]
            reverse_rhs = reversed(rule.rhs)
            reverse_counter = counter + len(reverse_rhs)
            for sym in reverse_rhs:
                if include_terminals or isinstance(sym, NonTerminal):
                    new_node = ParseTreeNode(reverse_counter, sym)
                    stack.push((sym, new_node))
                    children = [new_node] + children
                    counter += 1
                    reverse_counter -= 1
            top_node.children = children
            top_node.rule_number = rule_id
        return root

    def apply_actions(self, node, pass_all=True):
        #~ print "current env:" + str(dir(self.rule_actions_module))
        return self.__rec_apply_actions(node, pass_all)
    
    # Public mutators
    
    def __setstate__(self, dict):
        self.__dict__.update(dict)
        if type(self.rule_actions_module) is str:
            exec "import " + self.rule_actions_module
            self.rule_actions_module = eval(self.rule_actions_module)
    
    """
    Private instance methods
    """
    
    # Private accessors, selectors and extractors
    
    def __rec_apply_actions(self, node, pass_all=True, level=0):
        #~ print ' ' * level + "apply action to node id: "+str(node.identifier)+" symbol: "+str(node.symbol)
        if node.isleaf():
            #~ print ' ' * level + "leaf:",str(node.symbol)
            if isinstance(node.symbol, EmptyString):
                return node.symbol
            #~ print ' ' * level + "leaf type:",node.symbol.__class__.__name__
            #~ print ' ' * level + "leaf value type:",node.symbol.value().__class__.__name__
            return node.symbol.value()
        rule = self.cfg[node.rule_number]
        children_outcome = []
        for child in node.children:
            #~ print ' ' * level + "- looking at child id:"+str(child.identifier)+" symbol: "+str(child.symbol)
            if pass_all or isinstance(child.symbol, NonTerminal):
                #~ print ' ' * level + "- recursion"
                child_outcome = self.__rec_apply_actions(child, pass_all, level + 2)
                if not isinstance(child_outcome, EmptyString):
                    children_outcome.append(child_outcome)
        #~ print ' ' * level + "applying post-action:\n" + str(rule.action)
        #~ print ' ' * level + "- with args:"+str([str(v) for v in children_outcome])
        #~ print ' ' * level + "- with args:"+str([v for v in children_outcome])
        #~ print 'RA:',str(rule.action.__class__)
        if type(rule.action) is str:
            #~ print "rule action is str"
            action_name = rule.action
        elif isinstance(rule.action, RuleAction):
            #~ print "rule action is RuleAction "+str(rule.action.__class__)
            action_name = rule.action.name
        #~ print ' ' * level + "current env:" + str(dir(self.rule_actions_module))
        if self.rule_actions_module is None:
            #~ print "internal\n"+str(rule.action)
            action = rule.action
        elif type(self.rule_actions_module) is ModuleType:
            #~ print "from module"
            action = getattr(self.rule_actions_module, action_name)
        outcome = action(*children_outcome)
        #~ print ' ' * level + "result of recursive call: "+str(outcome)
        #~ print
        return outcome


class ParsingTable:
    
    def __init__(self, cfg, discard_choices=False, verbose=False):
        #~ print "Creating parsing table"
        self.cfg = cfg
        self.table = dict((n, dict((t.value(), set([])) \
                                   for t in cfg.terminals.union(SymbolSet([EmptyString()])))) \
                          for n in cfg.non_terminals)
        #~ print "Filling table"
        self.__fill_table()
        #~ print "Table filled"
        if discard_choices:
            #~ print "Discarding choices"
            self.__discard_choices(verbose)
            #~ print "Choices discarded"
        self.col_size = 15
    
    def __setitem__(self, nt_pair, value):
        self.table[nt_pair[0]][nt_pair[1]] = value
        
    def __getitem__(self, nt_pair):
        return self.table[nt_pair[0]][nt_pair[1]]
    
    def __fill_table(self):
        #~ print 'compute first sets'
        first_sets = self.__compute_first_sets()
        self.first_sets = first_sets
        #~ print 'compute follow sets'
        follow_sets = self.__compute_follow_sets(first_sets)
        self.follow_sets = follow_sets 
        #~ print 'sets computed'
        non_terminal_first_sets, word_first_sets = first_sets
        self.non_terminal_first_sets = non_terminal_first_sets
        self.word_first_sets = word_first_sets
        rule_number = 0
        for rule in self.cfg.rules:
            non_terminal = rule.lhs
            #~ non_terminal = rule.lhs.value()
            first_set_has_the_empty_string = False
            for symbol in word_first_sets[rule_number]:
                #~ terminal = symbol
                terminal = symbol.value()
                self.table[non_terminal][terminal].add(rule_number)
                if isinstance(symbol, EmptyString):
                    first_set_has_the_empty_string = True
            if first_set_has_the_empty_string:
                for symbol in follow_sets[non_terminal]:
                    #~ terminal = symbol
                    terminal = symbol.value()
                    self.table[non_terminal][terminal].add(rule_number)
            rule_number += 1
    
    def __discard_choices(self, verbose=False):
        for non_terminal in self.table:
            for terminal in self.table[non_terminal]:
                conflict = False
                this_set = self.table[non_terminal][terminal]
                if len(this_set) > 1:
                    conflict = True
                    if verbose:
                        print_time("* module: llparser.py")
                        print "  Warning: the parser generator spotted a possible conflict:"
                        print "  Rule conflict possible in the parsing table in the entry for "
                        print "    Non-terminal: ",str(non_terminal)
                        print "    Terminal:     ",str(terminal)
                        print "    Conflicting rules: ",str(this_set)
                        for r in this_set:
                            print "      " + str(r) + ": " + str(self.cfg.rules[r])
                        print "    Firsts set of "+str(non_terminal)+": "
                        print "      " + str([str(e) for e in self.first_sets[0][non_terminal]])
                        print "    Follow set of "+str(non_terminal)+": "
                        print "      " + str([str(e) for e in self.follow_sets[non_terminal]])
                try:
                    selection = this_set.pop()
                    if conflict and verbose:
                        print "  Choosing rule:"
                        print "    " + str(selection)
                    self.table[non_terminal][terminal] = selection
                except KeyError:
                    self.table[non_terminal][terminal] = None
    
    def __compute_first_sets(self):
        #~ print "Computing first sets"
        non_terminal_first_sets = dict((n, set([])) for n in self.cfg.non_terminals)
        word_first_sets = [set([]) for rule in self.cfg.rules]
        first_sets = (non_terminal_first_sets, word_first_sets)
        previous_sets = (None, None)
        #~ print "First sets ="
        #~ print_first_sets(first_sets)
        #~ print "Previous = "
        #~ print_first_sets(previous_sets)
        i = 0
        while first_sets != previous_sets:
            #~ print "ITERATION #"+str(i)+" ******************************************************"
            #~ print 'FIRST SETS'
            #~ print_first_sets(first_sets)
            #~ print 'PREVIOUS'
            #~ print_first_sets(previous_sets)
            previous_sets = deepcopy(first_sets)
            non_terminal_first_sets, word_first_sets = first_sets
            i+=1
            rule_number = 0
            for rule in self.cfg.rules:
                #~ print "rule: " + str(rule)
                #~ print
                word_first_set = self.__get_word_first_set(rule.rhs, first_sets)
                #~ print "word firsts = "
                #~ print_word_first_set(word_first_set)
                #~ print
                word_first_sets[rule_number] |= word_first_set
                rule_number += 1
            rule_number = 0
            for rule in self.cfg.rules:
                #~ print "  rule: " + str(rule)
                #~ print
                #~ #non_terminal_first_sets[rule.lhs.value()] |= word_first_sets[rule_number]
                non_terminal_first_sets[rule.lhs] |= word_first_sets[rule_number]
                rule_number += 1
            first_sets = (non_terminal_first_sets, word_first_sets)
            #~ print "******************************************************************\n"
        #~ print "First sets ="
        #~ print_first_sets(first_sets)
        #~ print "Previous = "
        #~ print_first_sets(previous_sets)
        #~ print "First sets computed"
        return first_sets

    
    def __get_word_first_set(self, word, first_sets):
        #~ print '=======> gwfs:',str([str(s) for s in word])
        if len(word) == 0:
            return set([EmptyString()])
            #~ return set([EmptyString()])
        elif isinstance(word[0], Terminal):
            return set([word[0]])
        elif isinstance(word[0], EmptyString) and len(word) == 1:
            return set([word[0]])
        non_terminal_first_sets, word_first_sets = first_sets
        non_terminal = word[0]
        #~ print '-- gwfs,nt:', str(non_terminal)
        #~ non_terminal = word[0].value()
        this_set = non_terminal_first_sets[non_terminal]
        #~ print '-- gwfs,ts:',str(set(str(e) for e in this_set))
        has_empty_string = False
        for element in this_set:
            if isinstance(element, EmptyString):
                has_empty_string = True
                break
        #~ print '-- gwfs,hes:',str(has_empty_string)
        if not has_empty_string:
            return this_set
        else:
            empty_removed = set(element for element in this_set \
                                if not isinstance(element, EmptyString))
            rest = self.__get_word_first_set(word[1:], first_sets)
            #~ print '-- gwfs,hes,er:',str(set(str(e) for e in empty_removed))
            #~ print '-- gwfs,hes,re:',str(set(str(e) for e in rest))
            result = empty_removed | rest
            #~ print '-- gwfs,hes,ans:',str(set(str(e) for e in result))
            return result
        
    def __get_word_first_set_with_empty_str(self, word, first_sets):
        #~ print 'gwfs:',str(word)
        if len(word) == 0:
            return set([])
            #~ return set([EmptyString()])
        elif isinstance(word[0], Terminal):
            return set([word[0]])
        elif isinstance(word[0], EmptyString) and len(word) == 1:
            return set([word[0]])
        non_terminal_first_sets, word_first_sets = first_sets
        non_terminal = word[0]
        #~ print '-- gwfs,nt:', str(non_terminal)
        #~ non_terminal = word[0].value()
        this_set = non_terminal_first_sets[non_terminal]
        #~ print '-- gwfs,ts:',str(set(str(e) for e in this_set))
        has_empty_string = False
        for element in this_set:
            if isinstance(element, EmptyString):
                has_empty_string = True
                break
        #~ print '-- gwfs,hes:',str(has_empty_string)
        if not has_empty_string:
            return this_set
        else:
            rest = self.__get_word_first_set_with_empty_str(word[1:], first_sets)
            #~ print '-- gwfs,hes,er:',str(set(str(e) for e in empty_removed))
            #~ print '-- gwfs,hes,re:',str(set(str(e) for e in rest))
            result = this_set | rest
            #~ print '-- gwfs,hes,ans:',str(set(str(e) for e in result))
            return result
    
    def __compute_follow_sets(self, first_sets):
        follow_sets = dict((n, set([])) for n in self.cfg.non_terminals)
        previous_sets = None
        while follow_sets != previous_sets:
            previous_sets = deepcopy(follow_sets)
            rule_number = 0
            for rule in self.cfg.rules:
                non_terminal = rule.lhs
                #~ non_terminal = rule.lhs.value()
                i = 0
                for symbol in rule.rhs:
                    if isinstance(symbol, NonTerminal):
                        rest = rule.rhs[i+1:]
                        followers = self.__get_word_first_set(rest, first_sets)
                        #~ followers = self.__get_word_first_set_with_empty_str(rest, first_sets)
                        #~ print '-->',str(symbol)
                        follow_sets[symbol] |= followers
                        #~ #follow_sets[symbol.value()] |= followers
                        followers_has_the_empty_string = False
                        for follower in followers:
                            if isinstance(follower, EmptyString):
                                followers_has_the_empty_string = True
                        if followers_has_the_empty_string or len(followers) == 0:
                            follow_sets[symbol] |= follow_sets[non_terminal]
                            #~ #follow_sets[symbol.value()] |= follow_sets[non_terminal]
                    i += 1
                rule_number += 1
        #~ print 'follow sets:'
        #~ print_follow_sets(follow_sets)
        #~ print '  ',str(dict((str(n), str(set(str(e) for e in follow_sets[n]))) for n in follow_sets))
        return follow_sets
    
    def __str__(self):
        s = " " * self.col_size + " "
        for t in self.cfg.terminals:
            token = t.value()
            if type(token) is ClassType:
                name = short_class_name(token)
            else:
                name = str(token)
            s += '|' + fit_str(name, self.col_size) + ' '
        s += '|' + str(EmptyString.symbol)
        s += '\n'
        s += '-' * (self.col_size + 1)
        s += ('+' + '-' * (self.col_size + 1)) * (len(self.cfg.terminals))
        s += '+' + '-' * max(self.col_size, len(str(EmptyString.symbol)))
        s += '\n'
        for n in self.cfg.non_terminals:
            s += fit_str(str(n.value()), self.col_size) + ' '
            for t in self.cfg.terminals:
                v = self.table[n][t.value()]
                if v is None:
                    v = '-'
                s += '|' + fit_str(str(v), self.col_size) + ' '
            last = self.table[n][EmptyString().value()]
            if last is None:
                last = '-'
            s += '|' + fit_str(str(last), self.col_size)
            s += '\n'
        return s


class ParseTreeNode:
    
    def __init__(self, identifier, symbol, rule_number=None, children=None):
        self.identifier = identifier
        self.symbol = symbol
        self.rule_number = rule_number
        if children is None:
            children = []
        self.children = children
        
    def isleaf(self):
        return self.children == []
    
    def add_child(self, node):
        #~ print "      adding "+str(node.identifier)+" to "+str(self.identifier)
        self.children.append(node)
        
    def __str__(self):
        children_str = ""
        for child in self.children:
            lines = str(child).split('\n')
            for line in lines:
                children_str += " " * 4 + line + "\n"
        return "{\n" \
            +  "  id: " + str(self.identifier) + "\n" \
            +  "  symbol: '" + str(self.symbol) + "'\n" \
            +  "  rule: " + str(self.rule_number) + "\n" \
            +  "  children: \n" + children_str \
            +  "}\n"
    
    def pprint(self):
        print "{\n" \
            +  "  id: " + str(self.identifier) + "\n" \
            +  "  symbol: '" + str(self.symbol) + "'\n" \
            +  "  rule: " + str(self.rule_number) + "\n" \
            +  "  children: " + str([(n.identifier, str(n.symbol)) for n in self.children]) \
            +  "}\n"


class ParsingException(Exception):
    def __init__(self, msg, token=None):
        self.msg = msg
        self.token = token
        self.linenum = token.linenum
        self.start_col = token.start_col
        self.end_col = token.end_col
        if token.end_col is None:
            self.end_col = token.start_col
    def pformat(self, line):
        r = "(ParsingException) \n"
        r += "line number: " + str(self.linenum) + "\n\n"
        r += line
        if self.start_col is not None:
            r += ' ' * self.start_col + '^' * (self.end_col - self.start_col + 1) + '\n'
        r += '\n'
        r += self.msg
        r += '\n'
        return r
    def pprint(self, line):
        print self.pformat(line)
    def __str__(self):
        return ('(ParsingException) '+str(self.msg))

# Private functions

def add_eoi(input_stream):
    eoi = EndOfInput()
    if type(input_stream) is str:
        input_stream += eoi.value()
    elif type(input_stream) is list:
        input_stream += [eoi.value()]
    elif type(input_stream) is GeneratorType:
        def addend(generator, v):
            for item in generator:
                yield item
            yield v
        input_stream = addend(input_stream, eoi)
    return input_stream
    
def position_in_text(string, index):
    if index >= len(string):
        raise IndexError()
    i = 0
    line_counter = 0
    in_line = 0
    while i < index:
        char = string[i]
        if char is '\n':
            line_counter += 1
            in_line = 0
        i += 1
    return (line_counter, in_line)

def fit_str(s, n, filling=' '):
    if len(s) >= n:
        return s[:n]
    return s + filling * (n - len(s))


import pprint
pp = pprint.PrettyPrinter()


# Debugging functions

def print_first_sets(fs):
    ntfs, wfs = fs
    if fs == (None, None):
        print "Empty"
        return
    printable_ntfs = dict((str(n), set(str(s) for s in ntfs[n])) for n in ntfs)
    print "Non-terminal first-sets: "
    pp.pprint(printable_ntfs)
    print
    print_word_first_sets(wfs)
    print

def print_word_first_sets(wfs):
    if wfs == None:
        print "Wmpty"
        return
    printable_wfs = [[str(s) for s in w] for w in wfs]
    print "Word first_sets"
    pp.pprint(printable_wfs)

def print_word_first_set(s):
    printable_s = set(str(e) for e in s)
    pp.pprint(printable_s)
def print_follow_sets(fos):
    fos = dict((str(n), set(str(e) for e in fos[n])) for n in fos)
    pp.pprint(fos)

def window(input_stream, current, size=5):
    #~ return input_stream[current: current + size]
    return current
        

def print_stack(stack):
    print "  stack = "
    for s, n in stack:
        if n is not None:
            print '    ' + str(s) + ',' + str(n.identifier)
        else:
            print '    ' + str(s) + ',' + str(n)
    