"""
Module: parser
Author: Ernesto Posse
Created on: June 22, 2006
Last update: Jan 22, 2007

Description:
  This module provides a factory function for creating parser objects, by giving the modules containing the CFG description and the rule actions
"""

import sys
import aperiot
sys.path.append(aperiot.__path__[0])
import os.path
sys.path.append(os.path.normpath(os.path.join(aperiot.__path__[0], '..')))

from aperiot.lexer import *
from aperiot.cfgs import *
from aperiot.llparser import *
from posse_tools.utils import *

# The speed of saving and loading parser objects depends on the pickle module.
# There are two alternative pickle modules in Python: pickle and cPickle
# The cPickle module is written in C and therefore is faster than the pickle
# module which is written in python itself. On the other hand, the cPickle module
# may suffer from some portability issues. If you want to improve the
# performance of saving and loading parsers, you can try uncommenting the following two lines...

import cPickle
pickle = cPickle

# ... and commenting out the following:

#~ import pickle

# The performance also depends on the pickle protocol (format) used. Currently
# There are 3 formats: 0, 1, and 2. Protocol 0 is an ASCII format which produces larger files and it takes longer to save and load, but is more portable. Protocols 1 ans 2 are binary formats which are faster to save and load but have some portability issues. Protocol 1 is meant for older versions of python. On the other hand, the ASCII protocol can be zipped with a very high compression rate.

ASCII      = 0
BINARY     = pickle.HIGHEST_PROTOCOL
OLD_BINARY = 1

formats = { ASCII: 'ASCII', BINARY: 'BINARY', OLD_BINARY: 'OLD_BINARY' }

#~ default_pickle_format = BINARY
default_pickle_format = ASCII


def build_parser(cfg_description_module, verbose=False):
    
    cfg_wat = cfg_description_module
    package_name = cfg_wat + '_cfg'
    full_module_name = package_name + '.' + cfg_wat
    exec "import " + full_module_name
    cfg_wat_module = eval(full_module_name)
    start = cfg_wat_module.start
    rules = cfg_wat_module.rules
    useindents = cfg_wat_module.useindents
    usenewlines = cfg_wat_module.usenewlines
    modules_to_import = cfg_wat_module.imports
    tokens = [getattr(cfg_wat_module, object_name) 
                for object_name in dir(cfg_wat_module) 
                if isinstance(getattr(cfg_wat_module, object_name), Token)]
    lexer = Lexer(tokens, 
                    yield_newlines=usenewlines, 
                    yield_indents=useindents)
    if verbose:
        print_time("* Lexer created for "+cfg_description_module)
    cfg = CFG(start, rules, modules_to_import)
    if verbose:
        print_time("* CFG created for "+cfg_description_module)
    cfg.normalize()
    if verbose:
        print_time("* CFG normalized for "+cfg_description_module)
    parser = Parser(cfg, lexer, rule_actions_module=None, verbose=verbose)
    if verbose:
        print_time("* Parser generated for "+cfg_description_module)
    return parser


def save_parser(parser, dir, module_name, pickle_format=default_pickle_format, verbose=False):
    current_dir = os.getcwd()
    target_dir = os.path.abspath(dir)
    os.chdir(target_dir)
    root, ext = os.path.splitext(module_name)
    package_name = root + '_cfg'
    cfg = parser.cfg
    modules_to_import = cfg.modules_to_import
    if not cfg.normalized:
        cfg.save_package(package_name=package_name, 
                            module='llk', 
                            modules_to_import=modules_to_import, 
                            additional_actions=[],
                            verbose=verbose)
    else:
        cfg.save_package(package_name=package_name, 
                            module='ll1', 
                            modules_to_import=modules_to_import, 
                            additional_actions=[],
                            verbose=verbose)
    #~ filename = os.path.abspath(os.path.join(package_name, root + '.pkl'))
    filename = os.path.join(package_name, root + '.pkl')
    #~ filename = root + '.pkl'
    file_handler = file(filename, 'w')
    pickle.dump(parser, file_handler, pickle_format)
    if verbose:
        print_time("* Parser saved in file '"+filename+"'")
        print_time("* File pickle format: "+formats[pickle_format])
    file_handler.close()
    os.chdir(current_dir)

def load_parser(module_name, verbose=False):
    root, ext = os.path.splitext(module_name)
    package_name = root + '_cfg'
    temp_table = {}
    exec "import " + package_name in temp_table
    package_path = eval(package_name, temp_table).__path__[0]
    filename = os.path.join(package_path, root + '.pkl')
    file_handler = file(filename, 'r')
    parser = pickle.load(file_handler)
    if verbose:
        print_time("* Parser loaded from file '"+filename+"'")
    file_handler.close()
    return parser




