"""
Module: path_setup
Author: Ernesto Posse
Created on: April 4, 2006
Last modified: April 5, 2006

Description:
    This module is used to ensure that a group of modules or packages is accessible, by searching for them in some search paths, and adding the necessary paths to sys.path so that they can be imported.
    
Typical use:
    In the main file of a project or package:

        import path_setup
        path_setup.setup_paths(required_modules=['program_entrypoint','road_runner','coyote'], strict=True, verbosity=1)

Constants:
    USUAL_SUSPECTS: str list = ['~']
        Contains a list of the typical directories to search for libraries.
    SAVED_PATHS_FILE: str = 'saved_paths.pth'
        File where found paths will be saved if cache flag is used in setup_paths

Functions:
    setup_paths(required_modules: str list, search_paths:str list=USUAL_SUSPECTS, strict:bool=False, cache=True, verbosity:int=0) -> None
        Searches for all required packages and modules in required_modules and adds the necessary paths to sys.path so that they can be imported. If the strict flag is True, a PathSetupError exception will be raised. If cache is true, the new paths will be saved in SAVED_PATHS_FILE on the current working directory. Note that the current directory is always included in the search paths.
        
    search_for_module(package_or_module_name: str, search_paths: str list verbosity:int=0) -> str
        Searchers for the package or module in the search paths and returns the directory where it is located, or None if it is not found.
        
Classes:
    PathSetupError(Exception)
        This class is used to report missing modules when using setup_path in strict mode.

History:
    April 4, 2006: Created. 
"""

#~ print "module path_setup"

import sys
import os
import os.path

# Having big directories such as '/usr/local/lib', '/usr/lib', '/usr/share', or '/usr/local/share' is a bad idea as searching these will considerably slow down the setup process.
USUAL_SUSPECTS = ['..']
SAVED_PATHS_FILE = 'saved_paths.pth'


# Public functions

def setup_paths(required_modules=[], search_paths=USUAL_SUSPECTS, strict=False, cache=True, verbosity=0):
    """
    setup_paths(required_modules: str list, search_paths:str list=USUAL_SUSPECTS, strict:bool=False, cache=True, verbosity:int=0) -> None
    
    Description:
        Searches for all required packages and modules in required_modules and adds the necessary paths to sys.path so that they can be imported. If the strict flag is True, a PathSetupError exception will be raised. If cache is true, the new paths will be saved in SAVED_PATHS_FILE on the current working directory. Note that the current directory is always included in the search paths.
    
    Parameters:
        required_modules: str list
            A list of module or package names
        search_paths: str list
            A list of paths to search
        cache: bool=True
            If true, saves the newly found paths to a file called 
        verbosity:int=0
            Level of verbosity. If non zero, print messages of the setup process.
    Returns:
        None
    """
    cwd = os.getcwd()
    if verbosity >= 1:
        print "(path_setup) Setting up paths"
        print "(path_setup) Current directory: '" + cwd + "'"
    if cwd not in sys.path:
        sys.path.append(cwd)
    search_paths = [os.path.abspath(os.path.expanduser(path)) for path in search_paths]
    existing_paths = load_paths(verbosity=verbosity)
    search_paths += existing_paths
    if cwd not in search_paths:
        search_paths.append(cwd)
    #~ search_paths = simplify_search_paths(search_paths)
    search_paths = rearrange_search_paths(search_paths)
    new_paths = []
    for item in required_modules:
        dir = search_for_module(item, search_paths, verbosity)
        if dir is not None:
            if dir not in sys.path:
                sys.path.append(dir)
            new_paths.append(dir)
        elif strict:
            raise PathSetupError(lib)
    new_paths += search_paths
    if cache:
        save_paths(new_paths, verbosity=verbosity)
    if verbosity >= 1:
        print "(path_setup) Path setup done"


def search_for_module(package_or_module_name, search_paths, verbosity=0):
    """
    search_for_module(package_or_module_name: str, search_paths: str list, verbosity:int=0) -> str
    
    Description:
        Searchers for the package or module in the search paths and returns the directory where it is located, or None if it is not found.
        
    Parameters:
        package_or_module_name: str
        search_paths: str list
        verbosity: int = 0
            Level of verbosity. If set to true, and the module is not found, a warning will be printed.
    
    Returns
        str
            The directory where package_or_module_name is located under search_paths, or None if not found.
    """
    for path in search_paths:
        dir = search_under(package_or_module_name, path, verbosity) 
        if dir is not None:
            return dir
    if verbosity >= 1:
        print "(path_setup) Warning: Module '" + package_or_module_name + "' not found. Without this module, some features may be unavailable, the program may fail to load, or it may crash."
    return None


# Public classes

class PathSetupError(Exception):
    def __init__(self, module):
        self.module = module
    def __str__(self):
        return "Module '" + str(self.module) + "' not found."


# Private functions

def dir_is_package(dir_name, root):
    full_dir = os.path.abspath(os.path.join(root, dir_name))
    return ('__init__.py' in os.listdir(full_dir))


def is_subdir_of(dir1, dir2):
    norm_dir1 = os.path.abspath(os.path.expanduser(dir1))
    norm_dir2 = os.path.abspath(os.path.expanduser(dir2))
    return os.path.commonprefix([norm_dir1, norm_dir2]) is norm_dir2 \
        and os.path.isdir(norm_dir1) and os.path.isdir(norm_dir2)


def load_paths(filename=SAVED_PATHS_FILE, verbosity=0):
    """
    load_paths(filename:str=SAVED_PATHS_FILE, verbosity:int=0) -> str list
    
    Description:
        Reads the file named filename and returns the list of paths in this file, if any. The file is expected to be a text file where each line is a path. Lines that are not paths are ignored.
    
    Parameters:
        filename:str=SAVED_PATHS_FILE
        verbosity:int=0
        
    Returns:
        str list
            The list of paths read from the file.
    """
    filename = os.path.join(os.getcwd(), filename)
    loaded_paths = []
    if verbosity >= 2:
        print "(path_setup) Loading paths from '" + filename + "'"
    if os.path.exists(filename):
        file = open(filename, 'r')
        lines = file.readlines()
        file.close()
        for line in lines:
            path_candidate = line.strip()
            if verbosity >= 3:
                print "(path_setup) Checking if '" + path_candidate + "' is a valid path."
            if os.path.exists(path_candidate):
                if verbosity >= 3:
                    print "(path_setup) Path is valid."
                loaded_paths.append(path_candidate)
            elif verbosity >= 3:
                print "(path_setup) Path '" + path_candidate + "' ignored: it doesn't exist or is not a valid path."
    if verbosity >= 3:
        print "(path_setup) Paths loaded"
    return loaded_paths
    

def rearrange_search_paths(search_paths=[]):
    """
    rearrange_search_paths(search_paths:str list=[]) -> str list
    
    Description:
        Reorganizes the list of paths so that deeper directories are placed first. Removes duplicates.
    
    Parameters:
        search_paths: str list  
            A list of paths (directories)
    Returns:
        str list
            A copy of the simplified list of paths
    """
    search_paths = [os.path.abspath(os.path.expanduser(path)) for path in search_paths]
    search_paths = list(set(search_paths))
    search_paths.sort(reverse=True)
    return search_paths


def save_paths(new_paths, filename=SAVED_PATHS_FILE, verbosity=0):
    """
    save_paths(new_paths: str list, filename:str=SAVED_PATHS_FILE, verbosity:int=0) -> None
    
    Description:
        Saves the list of paths to the file named filename in the current directory. It overwrites the existing file if any.
    
    Parameters:
        new_paths: str list
        filename:str=SAVED_PATHS_FILE
        verbosity:int=0
    
    Returns:
        None
    """
    filename = os.path.join(os.getcwd(), filename)
    if verbosity >= 2:
        print "(path_setup) Saving new paths in '" + filename + "'"
    if verbosity >= 3 and os.path.exists(filename):
        print "(path_setup) Overwriting file"
    file = open(filename, 'w')
    for path in new_paths:
        file.write(path + '\n')
        if verbosity >= 3:
            print "(path_setup) Writing path '" + path + "'"
    file.close()
    if verbosity >= 3:
        print "(path_setup) Paths saved"


def search_under(package_or_module_name, search_path=os.getcwd(), verbosity=0):
    """
    search_under(package_or_module_name:str, search_path:str=os.getcwd(), verbosity:int=0) -> str
    
    Description:
        Looks for the package or module in the search_path and all its subdirectories, and returns the exact subdirectory of the search_path where it is located (i.e. the "parent" or containing directory.)
    
    Parameters:
        package_or_module_name:str
        search_path:str=os.getcwd()
        verbosity:int=0
    
    Returns:
        str
            The path of the directory containing the given package_or_module_name.
    """
    search_path = os.path.abspath(os.path.expanduser(search_path))
    if verbosity >= 2:
        print "(path_setup) Searching for '" + package_or_module_name + "' under '" + search_path + "'"
    file_name = package_or_module_name + '.py'
    for root, dirs, files in os.walk(search_path):
        if file_name in files or package_or_module_name in dirs and dir_is_package(package_or_module_name, root):
            if verbosity >= 2:
                print "(path_setup) Module '" + package_or_module_name + "' found in '" + root + "'"
            return root

    
def simplify_search_paths(search_paths=[]):
    """
    simplify_search_paths:  str list -> str list
    
    Description: It returns a sorted copy of the given search_paths without duplicates and without directories that already have ancestors in the list.
    
    Parameters:
        search_paths: str list  
            A list of paths (directories)
    Returns:
        str list
            A copy of the simplified list of paths
    """
    search_paths = [os.path.abspath(os.path.expanduser(path)) for path in search_paths]
    search_paths.sort()
    simplified_paths = []
    if len(search_paths) > 0:
        simplified_paths = [search_paths[0]]
    for i in range(1, len(search_paths)):
        path = search_paths[i]
        is_subdir_of_previous = False
        for previous_path in search_paths[:i]:
            if is_subdir_of(path, previous_path):
                is_subdir_of_previous = True
                break
        if not is_subdir_of_previous:
            simplified_paths.append(path)
    return simplified_paths




