"""
Module: tries
Author: Ernesto Posse
Created on: April 12, 2006
Last modified: April 17, 2006

Description:
  Implements a trie data-structure.

Typical use:
    
    For tries without data:
    
        t = Trie(['ab', 'ac', 'abc'])
        try:
            value = t.search('ac')
        except KeyError, e:
            print 'not found'
    
    For tries with data:
    
        t = Trie({'ab':'data1', 'ac':'data2', 'abc':'data3'})
        data = t.search('ac')
        
        t.assign('ac', 'data4')    # changes the data associated to 'ac'
        t.assign('acd', 'data5')  # adds 'acd' to t and associates 'data5' to it.
    
    Trees with data can also be constructed with lists of tuples, or generator expressions:
    
        t = Trie([('ab',data1), ('ac', data2), ('abc', data3)])
    
    or
        t = Trie((k, data) for k in ['ab','ac',abc'])
    
    There is a mapping type syntax for these methods:
    
        t['ac']
    
    is equivalent to
    
        t.search('ac')
        
    and
    
        t['ac'] = 'data6'
        
    is equivalent to
    
        t.assign('ac', 'data6')
        
Classes:
    Trie:
        Methods:
            __init__(self: Trie, strings: (str, 'a) dict or str list or str set)
                Creates a new trie with the given strings. If the strings are a dictionary, the keys are expected to be the strings themselves and the values are the data to be stored at the end node for the corresponding string.
                
            assign(self: Trie, string: str, data: 'a=None) -> None
                assigns string to the trie. If data is none, associates the data with the string.
                
            search(self: Trie, string: str) -> 'a
                Looks for the string in the trie. It returns the data associated with the string, or raises a KeyError exception if the string is not in the trie.
            
            __str__(self: Trie) -> str
                Returns a string representation of the Trie, using the standard pretty-printer.

History:
    April 12, 2006: Created.
    April 17, 2006: Modified search to behave like dictionary search: on success it returns the data associated with the given key, and on failure it raises a KeyError exception.
"""

from types import *
import pprint

pp = pprint.PrettyPrinter()


class Trie:
    """
    Nodes in the tree are of the form
    
        [<id>, <arrows>, <data>]
    
    where <id> is a unique identifier within the trie (a natural number,)
    <arrows> is a dictionary with keys being the characters labelling the arrows from the node, and values being the target nodes, i.e. a pair (<key>,<value>) in this dictionary is really a pair (<char>, <target-node>), so there is an arrow from this node to the target node labelled <char>;
    and <data> is any assignitional data associated with this node.
    """
    
    def __init__(self, strings={}):
        self.strings = strings
        self.counter = 0
        self.root = [0, {}, None]
        if type(strings) is dict:
            for string in strings:
                self.assign(string, strings[string])
        else:
            for item in strings:
                if type(item) is str:
                    self.assign(item)
                elif type(item) is tuple or type(item) is list and len(item) >= 1:
                    self.assign(item[0], item[1:])
        self.pointer = self.root
    
    def __getitem__(self, string):
        return self.search(string)
    
    def __setitem__(self, string, new_data):
        self.assign(string, new_data)
    
    def assign(self, string, new_data=None):
        node = self.root
        arrows = node[1]
        new_node = None
        for char in string:
            if char in arrows:
                node = arrows[char]
                id, arrows, data = node 
            else:
                self.counter += 1
                node = [self.counter, {}, None]
                arrows[char] = node
                id, arrows, data = node
        node[2] = new_data
    
    def search(self, string):
        pointer = self.root[1]
        try:
            for char in string:
                id, pointer, data = pointer[char]
        except KeyError:
            raise KeyError(string)
        return data
    
    def reset(self):
        self.pointer = self.root
    
    def step(self, char):
        id, arrows, data = self.pointer
        new_node = arrows[char]
        self.pointer = new_node
        new_id, new_arrows, new_data = new_node
        return new_data
    
    #~ def incsearch(self, string):
        #~ node = self.root
        #~ pointer = node[1]
        #~ try:
            #~ for char in string:
                #~ node = pointer[char]
                #~ id, pointer, data = node 
                #~ yield (True, id,data)
        #~ except KeyError:
            #~ yield (False, id, data)
        #~ #yield node[0],node[2]
    
    #~ def old_search(self, string):
        #~ pointer = self.root[1]
        #~ try:
            #~ for char in string:
                #~ id, pointer, data = pointer[char]
        #~ except KeyError:
            #~ if type(self.strings) is DictType:
                #~ return (False, None)
            #~ else:
                #~ return False
        #~ if type(self.strings) is DictType:
            #~ return (True, data)
        #~ else:
            #~ return True
    
    #~ def old_step(self, char):
        #~ try:
            #~ id, arrows, data = self.pointer
            #~ new_node = arrows[char]
            #~ self.pointer = new_node
            #~ new_id, new_arrows, new_data = new_node
            #~ return new_data
        #~ except KeyError:
            #~ raise KeyError()
        
        
    def __str__(self):
        return pp.pformat(self.root)



