"""
Module: utils
Author: Ernesto Posse
Created on: Jun 9, 2006
Last modified: Dec 14, 2006

Description:
  This implement general utilities used throughout.
"""

import sys
import inspect
import traceback
import os.path
from threading import *
from time import *
from types import *

printer_lock = Lock()

class Verbose:
    
    def __init__(self, enabled=False, level=None, mask='all'):
        self.enabled = enabled
        self.verbosity = None
        self.set_verbose_level(level, mask)
        
    def set_verbose_level(self, level, mask='Ttmcfl'):
        self.one_only = 'o' in mask
        if mask == 'all':
            mask = 'Ttmcfl'
        self.mask = mask
        self.show_clock  = 'T' in mask
        self.show_thread = 't' in mask
        self.show_module = 'm' in mask
        self.show_class  = 'c' in mask
        self.show_func   = 'f' in mask
        self.show_lineno = 'l' in mask
        if level in ['', 'all']:
            self.verbosity = sys.maxint
        else:
            try:
                level = int(level)
                if level >= 0:
                    self.verbosity = level
                else:
                    self.verbosity = None
            except ValueError:
                self.verbosity = None
            except TypeError:
                self.verbosity = None

    def note(self, level, format, *args):
        if self.enabled and self.verbosity:
            if self.one_only and level == self.verbosity or not self.one_only and level <= self.verbosity:
                time      = self.show_clock and "%s " % strftime('%X') or ''
                frame     = inspect.currentframe().f_back
                table     = frame.f_locals
                module    = frame.f_code.co_filename
                module    = self.show_module and "%s " % os.path.splitext(os.path.basename(module))[0] or ''
                classname = 'self' in table and table['self'].__class__.__name__ or '-'
                classname = self.show_class and "%s " % classname or ''
                function  = self.show_func and "%s " % frame.f_code.co_name or ''
                lineno    = self.show_lineno and "%3d" % frame.f_lineno or ''
                thread    = self.show_thread and "%s " % currentThread().getName() or ''
                message   = format % args
                message   = "%s%s%s%s%s%s: %s\n" % (time, thread, module, classname, function, lineno, message)
                sys.stderr.write(message)
    
    def traceback(self, level=-1):
        if self.enabled and self.verbosity:
            if self.one_only and level == self.verbosity or not self.one_only and level <= self.verbosity:
                time      = self.show_clock and "%s " % strftime('%X') or ''
                stack     = '\n'.join(traceback.format_stack())
                message   = "%s:\n%s\n" % (time, stack)
                sys.stderr.write(message)
        

log = Verbose()



def all(S):
    for item in S:
        if item is False:
            return False
    return True

def any(S):
    for item in S:
        if item is True:
            return True
    return False

def unzip(list_of_pairs):
    if list_of_pairs == []:
        return ([],[])
    r = unzip(list_of_pairs[1:])
    return ([list_of_pairs[0][0]] + r[0], [list_of_pairs[0][1]] + r[1])

def print_time(msg, abstm=False):
    printer_lock.acquire()
    if abstm:
        print strftime('%X'), time(), clock(), str(msg) + ";"
    else:
        print strftime('%X'), str(msg) + ";"
    printer_lock.release()

def fit(s, n, filler=' '):
    s = str(s)
    ns = ''
    if len(s) < n:
        ns = s + filler * (n - len(s))
    else:
        ns = s[:n]
    return ns

def indent(text, n=4):
    lines = text.splitlines()
    lines = [' ' * n + line for line in lines]
    return '\n'.join(lines)

def getindentation(text):
    lines = text.splitlines()
    smallest = 1000
    for line in lines:
        i = 0
        while i < len(line) and line[i].isspace():
            i += 1
        if i < smallest:
            smallest = i
    return smallest

def dedent(text, n=4):
    lines = text.splitlines()
    n = getindentation(text)
    lines = [line[n:] for line in lines]
    return '\n'.join(lines)

def common_prefix(s1,s2):
    prefix = ''
    for c1, c2 in map(None, s1, s2):
        if c1 == c2:
            prefix += c1
        else:
            break
    return prefix

def postfix_difference(s1, s2):
    a, b = '', ''
    i = 0
    while i < len(s1) and i < len(s2):
        if s1[i] != s2[i]:
            break
        i += 1
    while i < len(s1) or i < len(s2):
        if i < len(s1):
            a += s1[i]
        if i < len(s2):
            b += s2[i]
        i += 1
    return a, b

def short_class_name(class_obj):
    full_name = str(class_obj)
    path = full_name.split('.')
    return path[len(path) - 1]

def short_str(obj, max_size=10):
    if type(obj) is ClassType:
        return short_class_name(obj)
    return str(obj)[:max_size]

def print_dict(d,indentation=0, key_width=15, type_width=15, value_width=20, value_delim='"', items='ktv'):
    for k in d:
        s = ''
        if k is not '__builtins__':
            if 'k' in items:
                s += fit(k,key_width,' ')+' : '
            if 't' in items:
                if type(d[k]) is InstanceType:
                    s += fit(d[k].__class__.__name__ + ' instance', type_width)
                elif type(d[k]) is ClassType:
                    s += fit('class ' + d[k].__name__, type_width)
                else:
                    s += fit(str(type(d[k]).__name__), type_width)
            if 'v' in items:
                s += " |-> " + value_delim + short_str(d[k], value_width) + value_delim 
            s = indent(s, indentation)
        else:
            if 'k' in items:
                s += fit(k,key_width,' ')+' : '
            if 't' in items:
                s += fit('module', type_width)
            if 'v' in items:
                s += ' |-> ...'
        print s


name_counter = 0

def new_name(base='x',existing=[],upper=False):
    global name_counter
    if upper:
        base = base.upper()
    new = base + '_' + str(name_counter)
    name_counter += 1
    while new in existing:
        new = base + '_' + str(name_counter)
        name_counter += 1
    return new

class NameGenerator:
    
    def __init__(self, base_name='x', names_to_avoid=None, upper=False):
        if names_to_avoid is None:
            names_to_avoid = {}
        self.names_to_avoid = dict([(name, None) for name in names_to_avoid])
        self.base_name = str(base_name)
        if upper:
            self.base_name = self.base_name.upper()
        self.last_index = 0
        
    def get_new_name(self, base=None, names_to_avoid=None, upper=False):
        if base is None:
            base = self.base_name
        if names_to_avoid is not None:
            for key in names_to_avoid:
                self.names_to_avoid[key] = None
        if upper:
            base = str(base).upper()
        new_index = self.last_index
        new_name = base + '_' + str(new_index)
        while new_name in self.names_to_avoid:
            new_index += 1
            new_name = base + '_' + str(new_index)
        self.last_index = new_index + 1
        return new_name
        
class IndexedNameGenerator:
    
    def __init__(self, base_name='x', names_to_avoid=None, upper=False):
        if names_to_avoid is None:
            names_to_avoid = {}
        self.names_to_avoid = dict([(name, None) for name in names_to_avoid])
        self.base_name = str(base_name)
        if upper:
            self.base_name = self.base_name.upper()
        self.last_index = 0
        
    def get_new_name(self, base=None, names_to_avoid=None, upper=False):
        if base is None:
            base = self.base_name
        if names_to_avoid is not None:
            for key in names_to_avoid:
                self.names_to_avoid[key] = None
        if upper:
            base = str(base).upper()
        new_index = self.last_index
        parts = base.split('_')
        try:
            last = int(parts[len(parts) - 1])
            base = '_'.join(parts[:len(parts)-1])
        except ValueError:
            last = 0
        new_index = last + 1
        new_name = base + '_' + str(new_index)
        while new_name in self.names_to_avoid:
            new_index += 1
            new_name = base + '_' + str(new_index)
        self.last_index = new_index + 1
        return new_name
        
